# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the formatters."""
import logging
import os
import unittest
import numpy as np
from pyretis.core import create_box, Particles, System
from pyretis.core.path import Path
from pyretis.core.pathensemble import PathEnsemble
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.inout.formats.formatter import (
    OutputFormatter,
    apply_format,
    format_number,
    get_log_formatter,
    PyretisLogFormatter,
    LOG_FMT,
    LOG_DEBUG_FMT,
)
from pyretis.inout.formats.cross import CrossFormatter
from pyretis.inout.formats.energy import EnergyFormatter, EnergyPathFormatter
from pyretis.inout.formats.order import OrderFormatter, OrderPathFormatter
from pyretis.inout.formats.path import PathExtFormatter, PathIntFormatter
from pyretis.inout.formats.snapshot import (
    SnapshotFormatter,
    adjust_coordinate,
)
from pyretis.inout.formats.pathensemble import PathEnsembleFormatter
from .help import (
    assert_equal_path_dict,
    create_external_path,
    turn_on_logging,
    set_up_system,
)
from .help import (
    create_test_paths,
    DATA_CROSS,
    DATA_CROSS2,
    CORRECT_CROSS,
    DATA_ENERGY,
    CORRECT_ENERGY,
    DATA_ORDER,
    CORRECT_ORDER,
    CORRECT_PATH_EXT,
    PATH_EXT_RAW,
)


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class OutputFormatterTest(unittest.TestCase):
    """Test that OutputFormatter work as intended."""

    def test_formatter_header(self):
        """Test that the header work as intended."""
        formatter = OutputFormatter('test-formatter', header=None)
        self.assertTrue(formatter.header is None)

        with self.assertRaises(AttributeError):
            OutputFormatter('test-file', header='Just a text header')

        txt = 'This is the header, it is very long indeed'
        formatter = OutputFormatter('test-file', header={'text': txt})
        self.assertEqual(formatter.header, txt)

        formatter = OutputFormatter('test-file',
                                    header={'text': txt, 'width': 10})
        self.assertEqual(formatter.header, txt)

        formatter = OutputFormatter(
            'test-file',
            header={'width': (10, 10),
                    'labels': ('Label1', 'Label2')},
        )
        self.assertEqual(formatter.header, '#   Label1     Label2')
        formatter.header = txt
        self.assertEqual(formatter.header, txt)

    def test_parse(self):
        """Test that the generic parser work for simple data."""
        formatter = OutputFormatter('test-formatter', header=None)
        raw = '1 2.0 3.0     4.123'
        data = formatter.parse(raw)
        correct = [1.0, 2.0, 3.0, 4.123]
        self.assertEqual(len(data), len(correct))
        for i, j in zip(data, correct):
            self.assertAlmostEqual(i, j)


class TestCrossFormatter(unittest.TestCase):
    """Test the CrossFormatter formatter."""

    def test_cross_formatter(self):
        """Test that we can format crossing data."""
        formatter = CrossFormatter()
        self.assertEqual(formatter.header, '#     Step  Int Dir')
        for i, datai in enumerate(DATA_CROSS):
            for line in formatter.format(i, datai):
                self.assertEqual(line, CORRECT_CROSS[i])

    def test_parse(self):
        """Test that we can parse crossing data."""
        formatter = CrossFormatter()
        for i, j in zip(CORRECT_CROSS, DATA_CROSS2):
            data = formatter.parse(i)
            self.assertTupleEqual(data, j)


class TestEnergyFormatter(unittest.TestCase):
    """Test the EnergFMT formatter."""

    def test_energy_file_formatter(self):
        """Test that we can format energy data."""
        formatter = EnergyFormatter()
        fields = ['vpot', 'ekin', 'etot', 'temp']
        self.assertEqual(
            formatter.header,
            ('#     Time      Potential        Kinetic          '
             'Total    Temperature')
        )
        for step, i in enumerate(DATA_ENERGY):
            data = {key: j for (key, j) in zip(fields, i)}
            for line in formatter.format(step, data):
                self.assertEqual(line, CORRECT_ENERGY[step])

    def test_energy_missing_data(self):
        """Test the energy formatting when some data is missing."""
        formatter = EnergyFormatter()
        data = {'vpot': 1.0, 'ekin': 2.0, 'etot': 3.0}
        for line in formatter.format(0, data):
            self.assertEqual(
                line,
                ('         0       1.000000       2.000000'
                 '       3.000000            nan')
            )


class TestEnergyPathFormatter(unittest.TestCase):
    """Test the EnergPathFMT formatter."""

    def test_energy_path_formatter(self):
        """Test that we can format energy data from paths."""
        formatter = EnergyPathFormatter()
        paths, _ = create_test_paths(npath=1)
        paths[0].generated = 'Cha-Cha'
        correct = ['# Cycle: 0, status: ACC, move: Cha-Cha',
                   '#     Time      Potential        Kinetic']
        for i, line in enumerate(formatter.format(0, [paths[0], 'ACC'])):
            if i < 2:
                self.assertEqual(line, correct[i])
            else:
                point = paths[0].phasepoints[i-2]
                txt = '{:>10d} {:>14.6f} {:>14.6f}'.format(
                    i-2, point.particles.vpot, point.particles.ekin
                )
                self.assertEqual(line, txt)
        # Test handling of empty paths:
        with self.assertRaises(StopIteration):
            next(formatter.format(0, [None, 'ACC']))
        # Just check what happens if we are missing one of the attributes:
        del paths[0].phasepoints[0].particles.vpot
        for i, line in enumerate(formatter.format(0, [paths[0], 'ACC'])):
            if i == 2:
                split = line.split()
                self.assertEqual('nan', split[1])


class TestOrderFormatter(unittest.TestCase):
    """Test the OrderFormatter formatter."""

    def test_order_parameter_formatter(self):
        """Test that we can format order parameter data."""
        formatter = OrderFormatter()
        self.assertEqual(formatter.header, '#     Time       Orderp')
        for i, datai in enumerate(DATA_ORDER):
            for line in formatter.format(i, datai):
                self.assertEqual(line, CORRECT_ORDER[i])


class TestOrderPathFormatter(unittest.TestCase):
    """Test the OrderPathFormatter formatter."""

    def test_order_path_formatter(self):
        """Test that we can format order parameter data for paths."""
        formatter = OrderPathFormatter()
        paths, _ = create_test_paths(npath=1)
        paths[0].generated = 'ld'
        correct = ['# Cycle: 0, status: ACC, move: ld',
                   '#     Time       Orderp']
        for i, line in enumerate(formatter.format(0, [paths[0], 'ACC'])):
            if i < 2:
                self.assertEqual(line, correct[i])
            else:
                point = paths[0].phasepoints[i-2]
                fmt = ' '.join(['{:>10d}'] + len(point.order)*['{:>12.6f}'])
                txt = fmt.format(i-2, *point.order)
                self.assertEqual(line, txt)
        # Test handling of empty paths:
        with self.assertRaises(StopIteration):
            next(formatter.format(0, [None, 'ACC']))


def generate_system_snapshots(ndim, length=10, npart=10):
    """Generate a set of system snapshots.

    Parameters
    ----------
    ndim : int
        The number of dimensions to consider.
    length : int, optional
        The number of snapshots to generate.
    npart : int, optional
        The number of particles to add.

    """
    box = create_box(low=[0.]*ndim, high=[10.]*ndim)
    system = System(box=box)
    system.particles = Particles(dim=ndim)
    for i in range(npart):
        name = 'Kr'
        if i % 2 == 0:
            name = 'Ar'
        system.add_particle(np.zeros(ndim), mass=1.0, name=name, ptype=0)
    for i in range(length):
        for j in range(npart):
            for k in range(ndim):
                system.particles.pos[j, k] = (i + 1) * (j + 1) * (k + 1)
                system.particles.vel[j, k] = (i + 0.5) * (j + 0.5) * (k + 0.5)
                if i % 2 == 0:
                    system.particles.vel *= -1
        yield system


def compare_traj_fmt_file(test, formatter, correct_file, ndim):
    """Compare formatted trajectory output with a file.

    Parameters
    ----------
    test : object like :py:class:`unittest.TestCase`.
        The object used for testing.
    formatter : object like :py:class:`.SnapshotFormatter`
        Formatter used to format the trajectory/systems.
    correct_file : string
        Path to the file with the correctly formatted output.
    ndim : int
        The number of dimensions to use (1, 2 or 3).

    Returns
    -------
    out : boolean
        True, if the data is correctly formatted.

    """
    correct = []
    with open(correct_file, 'r') as infile:
        for line in infile:
            correct.append(line.rstrip())
    j = 0
    for i, system in enumerate(generate_system_snapshots(ndim,
                                                         length=5)):
        for line in formatter.format(i, system):
            test.assertEqual(line, correct[j])
            j += 1


class TestSnapshotFormatter(unittest.TestCase):
    """Test the SnapshotFormatter formatter."""

    def test_traj_formatter_full(self):
        """Test that we can format trajectories."""
        formatter1 = SnapshotFormatter(write_vel=False, fmt='full')
        formatter2 = SnapshotFormatter(write_vel=True, fmt='full')
        cases = [
            (1, 'trajfmt_full_novel_1D.txt', formatter1),
            (2, 'trajfmt_full_novel_2D.txt', formatter1),
            (3, 'trajfmt_full_novel_3D.txt', formatter1),
            (1, 'trajfmt_full_vel_1D.txt', formatter2),
            (2, 'trajfmt_full_vel_2D.txt', formatter2),
            (3, 'trajfmt_full_vel_3D.txt', formatter2),
        ]
        for (ndim, filename, formatter) in cases:
            correct_file = os.path.join(HERE, filename)
            compare_traj_fmt_file(self, formatter, correct_file, ndim)

    def test_traj_formatter(self):
        """Test that we can format trajectories."""
        formatter1 = SnapshotFormatter(write_vel=False)
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.formats.snapshot',
                                 level='WARNING'):
                formatter1.parse('some text')
        formatter2 = SnapshotFormatter(write_vel=True)
        cases = [
            (1, 'trajfmt_novel_1D.txt', formatter1),
            (2, 'trajfmt_novel_2D.txt', formatter1),
            (3, 'trajfmt_novel_3D.txt', formatter1),
            (1, 'trajfmt_vel_1D.txt', formatter2),
            (2, 'trajfmt_vel_2D.txt', formatter2),
            (3, 'trajfmt_vel_3D.txt', formatter2),
        ]
        for (ndim, filename, formatter) in cases:
            correct_file = os.path.join(HERE, filename)
            compare_traj_fmt_file(self, formatter, correct_file, ndim)

    def test_adjust_coordinate(self):
        """Test that we can adjust coordinates."""
        # 1D:
        coord = np.array([10, ])
        coord_ = adjust_coordinate(coord)
        self.assertTrue(np.allclose(coord_, [[10., 0., 0.]]))
        # 1 particle, 1D:
        particles = Particles(dim=1)
        particles.add_particle(np.array([1.0]),
                               np.zeros(1),
                               np.zeros(1))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([1.0, 0.0, 0.0])))
        # 1 particle, 2D:
        particles = Particles(dim=2)
        particles.add_particle(np.array([1.0, 1.0]),
                               np.zeros(2),
                               np.zeros(2))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([1.0, 1.0, 0.0])))
        # 1 particle, 3D:
        particles = Particles(dim=3)
        particles.add_particle(np.array([1.0, 1.0, 1.0]),
                               np.zeros(3),
                               np.zeros(3))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([1.0, 1.0, 1.0])))
        # 2 particles, 1D:
        particles = Particles(dim=1)
        particles.add_particle(np.array([1.0]),
                               np.zeros(1),
                               np.zeros(1))
        particles.add_particle(np.array([-1.0]),
                               np.zeros(1),
                               np.zeros(1))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([[1., 0., 0.],
                                                   [-1., 0., 0.]])))
        # 2 particles, 2D:
        particles = Particles(dim=2)
        particles.add_particle(np.array([1.0, -1.0]),
                               np.zeros(2),
                               np.zeros(2))
        particles.add_particle(np.array([-1.0, 1.0]),
                               np.zeros(2),
                               np.zeros(2))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([[1., -1., 0.],
                                                   [-1., 1., 0.]])))

        # 3 particles, 3D:
        particles = Particles(dim=3)
        particles.add_particle(np.array([1.0, -1.0, 0.5]),
                               np.zeros(3),
                               np.zeros(3))
        particles.add_particle(np.array([-1.0, 1.0, -0.5]),
                               np.zeros(3),
                               np.zeros(3))
        pos = adjust_coordinate(particles.pos)
        self.assertTrue(np.allclose(pos, np.array([[1., -1., 0.5],
                                                   [-1., 1., -0.5]])))


class TestPathIntFormatter(unittest.TestCase):
    """Test that the PathFMT formatter work as intended."""

    def test_pathint_formatter(self):
        """Test the formatter for internal paths."""
        formatter = PathIntFormatter()
        paths, _ = create_test_paths(npath=2)
        k = -1
        for i, line in enumerate(formatter.format(0, [paths[1], 'ACC'])):
            if i == 0:
                self.assertEqual('# Cycle: 0, status: ACC', line)
            else:
                j = i - 1
                if j % 10 == 0:
                    k += 1
                    self.assertEqual(f'Snapshot: {k}', line)
                else:
                    # Assume 3D:
                    k_float = float(k)
                    pos = [k_float for _ in range(3)]
                    vel = [-k_float for _ in range(3)]
                    fmt = ' '.join(['{:15.9f}'] * 6)
                    txt = fmt.format(*pos, *vel)
                    self.assertEqual(txt, line)
        # Test handling of empty paths:
        with self.assertRaises(StopIteration):
            next(formatter.format(0, [None, 'ACC']))

    def test_parse(self):
        """Test that we can parse trajectory data."""
        formatter = PathIntFormatter()
        raw = '1.0 2.0 3.0 4.0 5.0 6.0'
        pos, vel = formatter.parse(raw)
        correct_pos = np.array([1., 2., 3.])
        correct_vel = np.array([4., 5., 6.])
        self.assertTrue(np.allclose(pos, correct_pos))
        self.assertTrue(np.allclose(vel, correct_vel))
        with self.assertRaises(ValueError):
            raw = '1.0'
            formatter.parse(raw)
        with self.assertRaises(ValueError):
            raw = '1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0'
            formatter.parse(raw)


class TestPathExtFormatter(unittest.TestCase):
    """Test that the external path formatter work as intended."""

    def test_pathext_formatter(self):
        """Test the formatter for external paths."""
        path, _ = create_external_path()
        formatter = PathExtFormatter()
        for corr, snap in zip(CORRECT_PATH_EXT,
                              formatter.format(0, [path, 'ACC'])):
            self.assertEqual(corr, snap)
        # Test handling of empty paths:
        with self.assertRaises(StopIteration):
            next(formatter.format(0, [None, 'ACC']))

    def test_parse(self):
        """Test parsing of path data."""
        formatter = PathExtFormatter()
        k = 0
        for line in CORRECT_PATH_EXT:
            if line.strip().startswith('#'):
                continue
            data = formatter.parse(line)
            self.assertListEqual(list(data), PATH_EXT_RAW[k])
            k += 1


class TestPathEnsembleFormatter(unittest.TestCase):
    """Test the formatting of path ensemble data."""

    def test_formatting(self):
        """Test the formatting of path data."""
        formatter = PathEnsembleFormatter()
        rgen = MockRandomGenerator(seed=0)
        path = Path(rgen)
        for i in range(10):
            path.append(
                set_up_system([i], np.ones((5, 3)) * i, np.ones((5, 3)) * i,
                              vpot=0.0, ekin=0.0, internal=True)
            )
        path.generated = ('sh', 0, 0, 0)
        ens = PathEnsemble(1, [0.5, 5.0, 8.0])
        ens.add_path_data(path, 'ACC')
        corrects = [
            ('         0          1          1 L M R      10 ACC sh'
             '  0.000000000e+00  9.000000000e+00       0       9'
             '  0.000000000e+00       0       0  1.000000000e+00'),
        ]
        for line, correct in zip(formatter.format(0, ens), corrects):
            self.assertEqual(line, correct)

    def test_parse(self):
        """Test parsing of formatted data."""
        formatter = PathEnsembleFormatter()
        raw = [
            ('         2          1          0 L * L    1019 '
             'NCR sh -9.020623174e-01 -3.052974085e-01    1018'
             '     554 -4.283683762e-01     721     292  2.000000000e+00'),
            ('         7          4          1 L M R    1140 '
             'ACC sh -9.018613917e-01  1.001603711e+00       0'
             '    1139 -2.765514723e-01     498     322  2.000000000e+00'),
        ]
        correct = [
            {
                'cycle': 2, 'generated': ['sh', -0.4283683762, 721, 292],
                'interface': ('L', '*', 'L'), 'length': 1019,
                'ordermax': (-0.3052974085, 554),
                'ordermin': (-0.9020623174, 1018), 'status': 'NCR',
                'weight': 2.
            },
            {
                'cycle': 7, 'generated': ['sh', -0.2765514723, 498, 322],
                'interface': ('L', 'M', 'R'), 'length': 1140,
                'ordermax': (1.001603711, 1139),
                'ordermin': (-0.9018613917, 0),
                'status': 'ACC',
                'weight': 2.
            },
        ]

        for i, j in zip(raw, correct):
            assert_equal_path_dict(formatter.parse(i), j)
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.formats.pathensemble',
                                 level='WARNING'):
                formatter.parse('1 2 3')
        # Also test a line with a comment:
        out = formatter.parse('1 2 3 A B C 4 ACC sh 5 6 7 8 9 8 7 2 # comment')
        cor = {'cycle': 1, 'generated': ['sh', 9.0, 8, 7],
               'interface': ('A', 'B', 'C'), 'length': 4, 'weight': 2.,
               'ordermax': (6.0, 8), 'ordermin': (5.0, 7), 'status': 'ACC'}

        assert_equal_path_dict(out, cor)

        # Also test a shorter (old version) line:
        out = formatter.parse('1 2 3 A B C 4 ACC sh 5 6 7 8 9 8 7')
        cor = {'cycle': 1, 'generated': ['sh', 9.0, 8, 7],
               'interface': ('A', 'B', 'C'), 'length': 4, 'weight': 1.,
               'ordermax': (6.0, 8), 'ordermin': (5.0, 7), 'status': 'ACC'}
        assert_equal_path_dict(out, cor)

        # Let it fail just because:
        out = formatter.parse('1 2 3 A B C 4 ACC sh 5 6 7 8 9 8 7')
        cor = {'cycle': 1, 'generated': ['sh', 9.0, 8, 7],
               'interface': ('A', 'B', 'C'), 'length': 4, 'weight': 2.,
               'ordermax': (6.0, 8), 'ordermin': (5.0, 7), 'status': 'ACC'}

        with self.assertRaises(AssertionError) as err:
            assert_equal_path_dict(out, cor)
        self.assertEqual(str(err.exception), 'Different weight')


class TestMethods(unittest.TestCase):
    """Test some of the methods defined in the formatters module."""

    def test_apply_format(self):
        """Test that we can apply a format."""
        txt = apply_format(12345.7, '{:7.2f}')
        self.assertEqual(txt, '1.2e+04')
        txt = apply_format(-1234568.9, '{:7.2f}')
        self.assertEqual(txt, ' -1e+06')
        txt = apply_format(-1234568.9, '{:8.2f}')
        self.assertEqual(txt, '-1.2e+06')
        txt = apply_format(-1234568.9, '{:9.2f}')
        self.assertEqual(txt, '-1.23e+06')
        txt = apply_format(123.45, '{:>10.2f}')
        self.assertEqual(txt, '    123.45')

    def test_format_number(self):
        """Test that we can format numbers as expected."""
        txt = format_number(99.9, 0.0, 100, fmtf='{0:<4.2f}',
                            fmte='{0:<4.2e}')
        self.assertEqual(txt, '99.90')
        txt = format_number(100.1, 0.0, 100, fmtf='{0:<4.2f}',
                            fmte='{0:<4.2e}')
        self.assertEqual(txt, '1.00e+02')
        txt = format_number(-100.1, 0.0, 100, fmtf='{0:<4.2f}',
                            fmte='{0:<4.2e}')
        self.assertEqual(txt, '-1.00e+02')

    def test_get_formatter(self):
        """Test that we can select the log formatter."""
        formatter = get_log_formatter(0)
        self.assertIsInstance(formatter, PyretisLogFormatter)
        self.assertEqual(formatter._fmt,  # pylint: disable=protected-access
                         LOG_DEBUG_FMT)

        formatter = get_log_formatter(100)
        self.assertIsInstance(formatter, PyretisLogFormatter)
        self.assertEqual(formatter._fmt,  # pylint: disable=protected-access
                         LOG_FMT)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package handles set-up of simulations from settings.

Package structure
-----------------

Modules
~~~~~~~

common.py (:py:mod:`pyretis.setup.common`)
    Common methods for handling settings. Defines a method to dynamically
    import methods and classes from user-specified modules.

createforcefield.py (:py:mod:`pyretis.setup.createforcefield`)
    Handle creation of force fields from input simulation settings.

createsimulation.py (:py:mod:`pyretis.setup.createsimulation`)
    Handle creation of simulations from input simulation settings.

createsystem.py (:py:mod:`pyretis.setup.createsystem`)
    Handle creation of systems from input simulation settings.

__init__.py
    This file. Handles imports for PyRETIS.

Important methods defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

create_engine (:py:func:`.create_engine`)
    Create an engine from input settings.

create_force_field (:py:func:`.create_force_field`)
    Create a force field from input settings.

create_orderparameter (:py:func:`.create_orderparameter`)
    Create an order parameter from input settings.

create_simulation (:py:func:`.create_simulation`)
    Create a simulation from input settings.

create_system (:py:func:`.create_system`)
    Create a system from input settings.
"""
from .common import create_orderparameter, create_engine
from .createsystem import create_system
from .createsimulation import create_simulation
from .createforcefield import create_force_field



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The order parameter for the hysteresis example."""
import logging
from pyretis.orderparameter.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class OrderX(OrderParameter):
    """A positional order parameter.

    Order parameter for the hysteresis example. In addition to using
    the position, we also use the energy to tell if we are in states A/B.


    Attributes
    ----------
    index : integer
        This is the index of the atom which will be used, i.e.
        system.particles.pos[index] will be used.
    inter_a : float
        An interface such that we are in state A for postions < inter_a.
    inter_b : float
        An interface such that we are in state B for postions > inter_b.
    energy_a : float
        An energy such that we are in state A for potential energy < energy_a.
    energy_b : float
        An energy such that we are in state A for potential energy < energy_b.
    dim : integer
        This is the dimension of the coordinate to use.
        0, 1 or 2 for 'x', 'y' or 'z'.
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the position or not.

    """

    def __init__(self, index, inter_a, inter_b, energy_a, energy_b,
                 dim='x', periodic=False):
        """Initialise the order parameter.

        Parameters
        ----------
        index : int
            This is the index of the atom we will use the position of.
        inter_a : float
            An interface such that we are in state A for postions < inter_a.
        inter_b : float
            An interface such that we are in state B for postions > inter_b.
        energy_a : float
            An energy such that we are in state A for
            potential energy < energy_a.
        energy_b : float
            An energy such that we are in state A for
            potential energy < energy_b.
        dim : string
            This select what dimension we should consider,
            it should equal 'x', 'y' or 'z'.
        periodic : boolean, optional
            This determines if periodic boundary conditions should be
            applied to the position.

        """
        txt = f'Position of particle {index} (dim: {dim})'
        super().__init__(description=txt)
        self.inter_a = inter_a
        self.inter_b = inter_b
        self.energy_a = energy_a
        self.energy_b = energy_b
        self.periodic = periodic
        self.index = index
        dims = {'x': 0, 'y': 1, 'z': 2}
        try:
            self.dim = dims[dim]
        except KeyError:
            msg = f'Unknown dimension {dim} requested'
            logger.critical(msg)
            raise

    def calculate(self, system):
        """Calculate the order parameter.

        Parameters
        ----------
        system : object like :py:class:`.System`
            This object is used for the actual calculation, typically
            only `system.particles.pos` and/or `system.particles.vel`
            will be used. In some cases `system.forcefield` can also be
            used to include specific energies for the order parameter.

        Returns
        -------
        out : float
            The order parameter.

        """
        particles = system.particles
        pos = particles.pos[self.index]
        lmb = pos[self.dim]
        if lmb < self.inter_a:
            if particles.vpot > self.energy_a:
                lmb = self.inter_a
        elif lmb > self.inter_b:
            if particles.vpot > self.energy_b:
                lmb = self.inter_b
        if self.periodic:
            return [system.box.pbc_coordinate_dim(lmb, self.dim)]
        return [lmb]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Compiler of PyRETIS simulation data.

This module is part of the PyRETIS library and can be used both for compiling
the simulation data into a compressed file and/or load the data for later
visualization.


Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Trajectory (:py:class:`.PathBase`)
    A base class to store trajectories composed by only orderp,
    collective variables and energies.

PathDensity (:py:class:`.PathBase`)
    A base class to assemble the data.

PathVisualize (:py:class:`.PathVisualize`)
    A base class to prepare for the visualization.


Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pyvisa_zip (:py:func: `.pyvisa_zip`)
    Compress the PyVisA output file in a .zip format.

pyvisa_unzip (:py:func: `.pyvisa_unzip`)
    Decompress the zip into PyVisA output.

remove_nan (:py:func: `.remove_nan`)
    Checks for the presence of nan values and replace them with a local,
    if available.

pyvisa_compress (:py:func: `.pyvisa_compress`)
    Compress PyRETIS outputs to a .hdf5 file.

"""
import os
import timeit
import warnings
import zipfile

import numpy as np
import pandas as pd
from scipy import stats

from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.pyvisa.common import where_from_to, get_cv_names

# Hard-coded labels for energies and time/cycle steps
ENERGYLABELS = ['time', 'cycle', 'potE', 'kinE', 'totE']


def pyvisa_zip(input_file):
    """Zip compress file of simulation data.

    Parameters
    ----------
    input_file : string
        The file to compress.

    """
    with zipfile.ZipFile(input_file + '.zip', 'w') as zipped_file:
        zipped_file.write(input_file, compress_type=zipfile.ZIP_DEFLATED)

    os.remove(input_file)


def pyvisa_unzip(origin, destination=None):
    """Unzip compressed file before load in visualizer.

    Parameters
    ----------
    origin : string
        Zipped file to unzip.
    destination : string, optional
        Unzipped file name.

    """
    msg = '###################################################\n'
    msg += '# File type recognized as `.zip`, unzipping to tmp file \n'
    msg += f'# {destination} before loading.\n'
    msg += '###################################################\n'
    print_to_screen(msg, level='message')
    with zipfile.ZipFile(origin) as zipped:
        zipped.extractall(path=os.path.dirname(os.path.abspath(origin)))
        # We here assume that only a file is stored in 'zipped'.
        if destination is not None:
            os.rename(zipped.namelist()[0], destination)


def remove_nan(data):
    """Remove nan from data.

    The function shall remove initial nan, assuming that they are originated
    by incomplete initial conditions (e.g. no energy file). In the case that
    nan appears as last cycle, it will not be fixed and an error shall rise
    up later in the code.

    Parameters
    ----------
    data : list, dict, object like pandas.DataFrame or pandas.Series
        Input data structure. If `np.nan` are present, they are
        replaced by the following entry. The method accounts for
        multiple consecutive `np.nan` occurrences.

    """
    nan = True
    inan = -1
    while nan:
        nan = False
        if type(data) in (dict, pd.DataFrame):
            for keys in data:
                if isinstance(data, pd.DataFrame):
                    # Pandas 2.0+ Copy-on-Write: data[keys] returns a copy,
                    # so we must modify a copy and assign it back explicitly.
                    col = data[keys].copy()
                    remove_nan(col)
                    data[keys] = col
                else:
                    remove_nan(data[keys])
            break
        for idx, data_point in reversed(list(enumerate(data))):
            if type(data_point) in (list, np.ndarray):
                remove_nan(data_point)
            else:
                # Check if data_point is a numeric type before calling np.isnan
                if (isinstance(data_point,
                               (int, float, np.floating, np.integer))
                        and np.isnan(data_point)):
                    nan = True
                    inan = idx
                    break
        if nan and inan == len(data) - 1:
            nan = False
        if nan and isinstance(data, pd.Series):
            data.iloc[inan] = data.iloc[inan + 1]
        elif nan:
            data[inan] = data[inan + 1]


def pyvisa_compress(runpath, input_file, pyvisa_dict):
    """Compress simulation data.

    Parameters
    ----------
    runpath : string
        The execution folder where the input files are.
    input_file : string
        The input file for compression.
    pyvisa_dict : dict
        It determines the section of pyvisa to use.

    """
    assert input_file, 'No input file'

    if input_file.endswith(('hdf5', 'zip')):
        msg = 'Cannot compress an already compressed file.'
        raise ImportError(msg)
    p_data = PathDensity(runpath, input_file)
    p_data.walk_dirs(only_ops=pyvisa_dict['only_order'])
    p_data.hdf5_data()


class Trajectory:
    """Class representing a simulation trajectory.

    This class defines the trajectories from the completed
    simulations, with all information available. The labels of
    the order parameter and collective variables will either be
    labeled in the fashion of opx, or from the names in the input file
    if the number of names given and number of descriptors in the system
    are equal.
    """

    def __init__(self, frames, info):
        """Initialize the class.

        Parameters
        ----------
        frames : pandas Dataframe or dict
            Dataframe/dict containing order parameter and energy-data
            for the trajectory. It contains:

            * `OP1`: string
              Order parameter.
            * `OP2`: string
              Collective variables, all other CV's will be named in
              increasing fashion, OP3, OP4 etc.
            * `PotE`: string
              Potential energy.
            * `KinE`: string
              Kinetic energy.
            * `TotE`: string
              Total energy.

        info : dict
            Dictionary containing information about the trajectory.
            It contains:

            * `ensemble_name`: string
              The name of the ensemble.
            * `cycle`: integer
              The cycle number.
            * `status`: string
              Status of accepted/rejected etc.
            * `MC-move`: string
              Generation move of the trajectory.
            * `MC-start`: string
              Generation starting point of the trajectory.
            * `ordermax`: float
              Max value of OP.
            * `ordermin`: float
              Min value of OP.
            * `stored`: boolean
              True if the trajectory has existing trajectory files.

        """
        self.info = info
        self.frames = frames


class PathDensity:
    """Perform the path density analysis.

    This class defines the path density analysis for completed simulations
    with several order parameters.

    """

    def __init__(self, basepath='.', iofile=None):
        """Initialize the class.

        Parameters
        ----------
        basepath : string, optional
            The path to the input file.
        iofile : string, optional
            The input file.

        Attributes
        ----------
        traj_dict : dict
            Values of order params and energy in all ensembles and
            info about trajectories.
            To each key, ensemble_name (e.g. 000, 001, etc.)
            the value is the list of respective Trajectory objects.

        infos : dict
            Information about the simulation,
            it contains:

            * `ensemble_names`: list
              List of ensemble names.
            * `interfaces`: list
              List of interface positions.
            * `num_op`: integer
              Number of order parameters.
            * `op_labels`: list
              List of order parameter names.
            * `energy_labels`: list
              List of energy entry labels.

        """
        self.basepath = os.path.abspath(basepath)

        if isinstance(iofile, PathDensity):
            self.traj_dict = iofile.traj_dict
            self.infos = iofile.infos
            return

        self.iofile = os.path.join(self.basepath, iofile)

        if iofile.endswith('.hdf5'):
            self.initialize_compressed(iofile)
            return

        # Getting interfaces from iofile
        settings = parse_settings_file(self.iofile)

        interfaces = settings.get('simulation', {}).get('interfaces', [])
        assert interfaces, 'Input file does not contain interface setting'

        intnames = ['0$^{-}$', '0$^{+}$']
        for i in range(1, len(interfaces) - 1):
            intnames.append(str(i) + '$^{+}$')
        path = []
        # Getting ensembles/folders from directory
        for fol in sorted(os.listdir(basepath)):
            if os.path.isdir(os.path.join(basepath, fol)) and fol.isdigit():
                path.append(fol)

        assert path, 'No files to analyse.'

        # Getting order parameters from order-file of first folder in path.
        # Careful here because the order parameter function and/or the
        # collective variables ones can produce a LIST. In this case the
        # counting will not correspond.
        filename = os.path.join(basepath, path[0], 'order.txt')
        with open(filename, 'r', encoding='utf-8') as figa:
            lines = len(figa.readlines())

        assert lines > 3, f'File order.txt in {path[0]} empty.'

        # Crash-proof, num_op is the mode of length of the last lines.
        with open(os.path.join(basepath, path[0], 'order.txt'),
                  encoding='utf-8') as temp:
            lasts = temp.readlines()[-10:]
            num_op = int(stats.mode([len(i.split()) for i in lasts])[0]) - 1

        op_names = get_cv_names(settings)
        op_labels = op_names if len(op_names) == num_op else \
            [f'op{i}' for i in range(1, num_op + 1)]

        self.traj_dict = {}
        self.order_parameter = op_labels[0]
        self.infos = {'ensemble_names': path,
                      'interfaces': interfaces,
                      'intf_names': intnames,
                      'num_op': num_op,
                      'op_labels': op_labels,
                      'energy_labels': ENERGYLABELS}

    def walk_dirs(self, only_ops=False):
        """Create a lists in acc or rej dictionary for all order parameters.

        First generate list of folders/ensembles to iterate through.
        Then search for number of order parameters(columns) in file in one
        of the folders of path, and create lists in acc/rej dictionaries
        for all order parameters.

        Lastly iterate through all folders and files, filling in correct data
        to the lists and dictionaries.

        Parameters
        ----------
        only_ops : boolean, optional
            If true, PathDensity will not collect data from energy files.

        """
        tic = [timeit.default_timer(), None]
        print_to_screen('###################################################',
                        level='message')
        print_to_screen('# PathDensity performing "walk" in \n# ' +
                        f'{os.getcwd()}/',
                        level='message')
        print_to_screen('# Number of subfolders (0**) = ' +
                        str(len(self.infos['ensemble_names'])),
                        level='message')
        print_to_screen((f'# Found {self.infos["num_op"]} order parameters in '
                         'output'),
                        level='message')
        print_to_screen('###################################################'
                        + '\n', level='message')

        # Looping over folders, reading energy and orderP
        cmax, cmin = 0, 0
        for ensemble_name in self.infos['ensemble_names']:

            tic[1] = timeit.default_timer()
            print_to_screen(f'Reading data from {ensemble_name}',
                            level='message')
            self.get_traj_op(ensemble_name)
            if not only_ops:
                self.get_traj_energy(ensemble_name)

            # Sort traj_dict by cycle
            self.traj_dict[ensemble_name] = \
                dict(sorted(self.traj_dict[ensemble_name].items()))
            for cyc in self.traj_dict[ensemble_name].keys():
                traj = self.traj_dict[ensemble_name][cyc]
                if int(cyc) > cmax:
                    cmax = cyc
                if self.order_parameter in traj.frames.columns:
                    traj.info['ordermax'] = np.max(
                        traj.frames[self.order_parameter])
                    traj.info['ordermin'] = np.min(
                        traj.frames[self.order_parameter])
                    traj.info['length'] = len(traj.frames.index)
            line = ('Done with folder, time used: '
                    f'{timeit.default_timer() - tic[1]:4.4f}s, proceeding.\n')
            print_to_screen('=' * len(line) + '\n' + line, level='success')

        self.infos['cycles'] = [cmin, cmax]

        print_to_screen('###################################################',
                        level='success')
        print_to_screen('# Data successfully retrieved, in cycles:',
                        level='success')
        print_to_screen((f'# {self.infos["cycles"][0]} to '
                         f'{self.infos["cycles"][1]}'),
                        level='success')
        print_to_screen(
            f'# Time spent: {timeit.default_timer() - tic[0]:.2f}s',
            level='success'
        )
        print_to_screen('###################################################'
                        + '\n', level='success')

    def hdf5_data(self):
        """Compress the data to a .hdf5 file."""
        print_to_screen('###################################################',
                        level='message')
        print_to_screen('# Compress dictionaries to file', level='message')

        data = pd.Series(self.traj_dict)
        infos = pd.Series(self.infos)

        pfile = 'pyvisa_compressed_data.hdf5'
        warnings.simplefilter(action='ignore',
                              category=pd.errors.PerformanceWarning)
        data.to_hdf(pfile, key='data', mode='w')
        infos.to_hdf(pfile, key='infos')
        print_to_screen(f'# {pfile}', level='message')
        pyvisa_zip(pfile)
        print_to_screen(f'# {pfile}.zip', level='message')
        print_to_screen('###################################################'
                        + '\n', level='message')

    def get_traj_op(self, ensemble_name):
        """Read order.txt files and collects data OP data.

        Parameters
        ----------
        ensemble_name : string

        Updates
        -------
        traj_dict : dict
            Dictionary containing Trajectory objects from simulation
            with filled in OP-data.

        """
        bad_characters = ["(", ")", ",", '"', "'"]
        self.traj_dict[ensemble_name] = {}
        weight = []
        statw = []
        frames = {}
        cycle = -1
        info = {}
        with open(os.path.join(self.basepath, ensemble_name, 'order.txt'),
                  'r', encoding='utf-8') as order:
            for idx, oline in enumerate(order):
                if oline.startswith('#') and 'Cycle' not in oline:
                    continue
                if oline.startswith('#') and 'Cycle' in oline:
                    if idx > 1:
                        self.fill_op(frames, info, ensemble_name, cycle)

                    text = oline.split()
                    cycle = int(text[2].rstrip(','))
                    frames = {}
                    info = {'ensemble_name': ensemble_name,
                            'cycle': cycle,
                            'status': text[4].rstrip(','),
                            'stored': False}

                    # Check for stored trajectory-files
                    if info['status'] == 'ACC':
                        status_line = 'traj-acc'

                        # Append statistical weights
                        statw.append(1)
                        weight.append(1)
                    else:
                        status_line = 'traj-rej'
                        # Append statistical weights
                        if weight:
                            weight[-1] += 1
                        statw.append(0)

                    path_to_cycles = (
                        f"{os.getcwd()}/{ensemble_name}/traj/{status_line}"
                    )

                    info['stored'] = os.path.exists(path_to_cycles +
                                                    f'/{cycle}/traj')

                    if len(text) > 6:  # For backward comparability
                        # Add MC move
                        info['MC-move'] = ''.join(filter(
                            lambda i: i not in bad_characters, text[6]))
                else:
                    # Read data line
                    odata = oline.split()
                    if not odata:
                        continue
                    frames[odata[0]] = {self.order_parameter: np.nan}

                    # Fill in OP-data to the frame
                    for label, data in zip(self.infos['op_labels'], odata[1:]):
                        frames[odata[0]][label] = np.array(data)

            self.fill_op(frames, info, ensemble_name, cycle)

        # Give weights to the trajectories
        for cycle in self.traj_dict[ensemble_name].keys():
            traj = self.traj_dict[ensemble_name][cycle]
            index = list(self.traj_dict[ensemble_name].keys()).index(cycle)
            if statw[index] == 1:
                traj.info['weight'] = weight[0]
                weight.pop(0)
            else:
                traj.info['weight'] = 0

    def fill_op(self, frames, info, ensemble_name, cycle):
        """Fill in OP-data to a trajectory.

        Function that fills the dictionary traj_dict containing
        Trajectory objects with data from the order parameter
        and the collective variables from the simulation.

        Parameters
        ----------
        frames: dict
            Dictionary containing energy-data for the trajectory.
        info : dict
            Information about the trajectory.
        ensemble_name : string
            The name of the ensemble.
        cycle : integer
            Cycle number.

        Updates
        -------
        traj_dict : Trajectory object
            Trajectory object with filled in OP-data.

        """
        orderp = self.order_parameter
        frames = pd.DataFrame.from_dict(
            frames, orient='index', dtype=float)
        for var in frames.columns:
            remove_nan(frames[var])
        if orderp in frames.columns and \
                frames[orderp].isnull().sum() != len(frames[orderp].index):
            start_end = where_from_to(np.array(frames[orderp]),
                                      self.infos['interfaces'][0],
                                      self.infos['interfaces'][-1])

            info['reactive'] = start_end[0] + start_end[1] == 'LR'
            self.traj_dict[ensemble_name][cycle] = Trajectory(frames, info)

    def get_traj_energy(self, ensemble_name):
        """Read energy.txt files and collects energy-data.

        Function that fills the dictionary traj_dict containing
        Trajectory objects with energy data from the simulation.

        Parameters
        ----------
        ensemble_name : string
            The name of the ensemble.

        Updates
        -------
        traj_dict : dict
            Dictionary containing Trajectory objects from simulation
            with filled in energy-data.

        """
        en_file = os.path.join(self.basepath, ensemble_name, 'energy.txt')
        if not os.path.exists(en_file):
            return
        cycle = 0
        eframes = {}
        with open(en_file, 'r', encoding='utf-8') as energy:
            for idx, eline in enumerate(energy):

                if eline.startswith('#') and 'Time' in eline:
                    continue

                if eline.startswith('#') and 'Cycle' in eline:
                    if idx > 1:
                        # Fill in energy data to the Trajectory
                        if cycle in self.traj_dict[ensemble_name].keys():
                            self.fill_energy(
                                self.traj_dict[ensemble_name][cycle],
                                eframes, ensemble_name, cycle)

                    cycle = int(eline.split()[2].rstrip(','))
                    eframes = {}

                else:
                    edata = eline.split()
                    if not edata:
                        continue
                    eframes[edata[0]] = {}

                    # in case of missing lines in energy.txt
                    if len(edata) < 3:
                        eframes[edata[0]]['potE'] = np.nan
                        eframes[edata[0]]['kinE'] = np.nan
                        continue
                    eframes[edata[0]]['potE'] = float(edata[1])
                    eframes[edata[0]]['kinE'] = float(edata[2])

            # End of file
            # Fill in energy data to the Trajectory
            if cycle in self.traj_dict[ensemble_name].keys():
                self.fill_energy(self.traj_dict[ensemble_name][cycle],
                                 eframes, ensemble_name, cycle)

    def fill_energy(self, traj, eframes, ensemble_name, cycle):
        """Fill in energy-data to a trajectory.

        Parameters
        ----------
        traj: Trajectory object
            Trajectory object to be filled in energy-data.
        eframes: dict
            Dictionary containing energy-data for the trajectory.
        ensemble_name : string
            The name of the ensemble.
        cycle : integer
            Cycle number.

        Updates
        -------
        traj_dict : Trajectory object
            Trajectory object with filled in energy-data.

        """
        if not eframes:
            return
        energy_frame = pd.DataFrame.from_dict(
            eframes, orient='index', dtype=float)
        for energy_term in energy_frame.columns:
            # If dataframe is only nan
            if energy_frame[energy_term].isnull().sum() == \
                    len(energy_frame[energy_term].index):
                return
            traj.frames[energy_term] = energy_frame[energy_term]
            remove_nan(traj.frames[energy_term])
        traj.frames['totE'] = \
            pd.to_numeric(traj.frames['potE']) + \
            pd.to_numeric(traj.frames['kinE'])
        self.traj_dict[ensemble_name][cycle] = traj

    def initialize_compressed(self, input_file):
        """Load PathDensity from a compressed file.

        Parameters
        ----------
        input_file : string
            The input file.

        """
        in_file = os.path.join(self.basepath, input_file)
        if os.path.isfile(in_file) and input_file.endswith('hdf5'):
            self.traj_dict = pd.read_hdf(in_file, key='data')
            self.infos = pd.read_hdf(in_file, key='infos')
        else:
            raise FileNotFoundError(f'File {in_file} not valid.')


class PathVisualize:
    """Class to define the visualization of data with PathDensity.

    Class definition of the visualization of data gathered from simulation
    directory using the PathDensity class.

    """

    def __init__(self, basepath='.', pfile=None):
        """Initialize the PathVisualize class.

        If a supported compressed input file is present, loads the pre-compiled
        data from it. Else, must use specific functions explicitly.

        Parameters
        ----------
        basepath : string, optional
            The path of the input file.
        pfile : string, optional
            The input file.

        """
        self.infos, self.traj_data = {}, {}
        self.op_labels = []
        self.x_list, self.y_list, self.z_list = [], [], []
        self.data_origin = []
        if pfile is None:
            self.pfile = None
        elif pfile.endswith('rst'):
            self.pfile = pfile
        else:
            self.pfile = os.path.join(basepath, pfile)
            self.load_whatever()

    def load_whatever(self):
        """Load all possible supported files.

        This functions directs traffic towards the real loaders.
        Essentially, it does almost nothing.

        """
        assert os.path.isfile(self.pfile), f'{self.pfile} does not exist.'
        tmp = None
        if self.pfile.endswith('.zip'):
            origin = self.pfile
            self.pfile = self.pfile.rstrip('.zip')
            # If a zip file has been loaded through the file menu
            tmp = os.path.join(os.path.dirname(os.path.abspath(origin)),
                               os.path.splitext(self.pfile)[0] + '.tmp' +
                               os.path.splitext(self.pfile)[1])
            self.pfile = tmp
            pyvisa_unzip(origin, tmp)
        if self.pfile.endswith('.hdf5'):
            self.load_hdf5()
            if tmp:
                os.remove(tmp)
        else:
            raise ValueError(f'Format of {self.pfile} not recognised')

    def load_hdf5(self):
        """Load precompiled data from a hdf5 file.

        Function that loads precompiled data from a .hdf5 file made
        using pandas.

        """
        self.traj_data = pd.read_hdf(self.pfile, key='data')
        self.infos = pd.read_hdf(self.pfile, key='infos')
        self.op_labels = self.infos['op_labels']

    def load_traj(self, criteria):
        """Load relevant data from Trajectories.

        Parameters
        ----------
        criteria : dict
            Dictionary of the selection criteria for which data to load.
            It contains:

            * `x`: string
              Name of parameter to plot.
            * `y`: string
              Name of parameter to plot.
            * `z`: string
              Name of parameter to plot.
            * `ensemble_name`: string
              Name of ensemble to loop through.
            * `cycles`: tuple
              Cycles to loop over.
            * `status`: string
              Status of the path: accepted/rejected.
            * `MC-move`: string, optional
              Generation move of the trajectory.
            * `stored`: bool, optional
              True if the trajectory has available trajectory files.
            * `weights`: bool, optional
              Option to apply statistical weights to the trajectories,
              used in the weighted density plot.

        Returns
        -------
        x_list : list
            List of data from chosen parameter.
        y_list : list
            List of data from chosen parameter.
        z_list : list
            List of data from chosen parameter, if required.
        data_origin: list
            List of ensemble_name and cycle for each point, if required.

        """
        x_list, y_list, z_list, data_origin = [], [], [], []
        for cyc in range(criteria['cycles'][0], criteria['cycles'][1] + 1):
            if cyc in self.traj_data[criteria['ensemble_name']].keys():
                traj = self.traj_data[criteria['ensemble_name']][cyc]
                # If the trajectory is empty
                if 'length' not in traj.info.keys():
                    continue  # pragma: no cover

                stat_weight = traj.info['weight'] if criteria.get('weight',
                                                                  False) else 1

                for key in criteria.keys():
                    # For different frequencies of outputs
                    if key in ['x', 'y', 'z']:
                        if criteria[key] not in ['time', 'cycle', 'None'] \
                                and criteria[key] not in traj.frames.columns:
                            break  # pragma: no cover
                    if key in traj.info.keys():
                        if key == 'status':
                            if criteria[key] == 'BOTH':
                                continue
                            if criteria[key] == 'REJ' and \
                                    traj.info[key] != 'ACC':
                                continue
                        elif ((key == 'stored' and not criteria[key]) or
                              (key == 'MC-move' and criteria[key] == 'All') or
                              (key in ['x', 'y', 'z', 'cycles', 'weight'])):
                            continue

                        if criteria[key] != traj.info[key]:
                            break
                else:
                    for _ in range(stat_weight):
                        data_origin.extend([[traj.info['ensemble_name'],
                                             traj.info['cycle']]] *
                                           traj.info['length'])
                        for xyz in {'x': x_list,
                                    'y': y_list,
                                    'z': z_list}.items():

                            if criteria[xyz[0]] == 'time':
                                xyz[1].extend(range(0, traj.info['length']))
                            elif criteria[xyz[0]] == 'cycle':
                                xyz[1].extend([traj.info['cycle']] *
                                              traj.info['length'])
                            elif criteria[xyz[0]] == 'None':
                                xyz[1].extend([1] * traj.info['length'])
                            else:
                                xyz[1].extend(
                                    traj.frames[criteria[xyz[0]]])
        return x_list, y_list, z_list, data_origin



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A class for writing text data for the analysis.

This module defines a text plotter which supports the same method as
the generic plotter, however, the output is human-readable text.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TxtPlotter (:py:class:`.TxtPlotter`)
    A class for writing text output.
"""
import gzip
import logging
import os
import shutil
import numpy as np
from pyretis.inout.plotting.plotting import Plotter
from pyretis.inout.common import name_file
from pyretis.inout.common import (ENERFILES, ENERTITLE, FLUXFILES,
                                  ORDERFILES, PATHFILES, PATH_MATCH)
from pyretis.inout.formats.txt_table import txt_save_columns

np.set_printoptions(legacy='1.21')

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['TxtPlotter']


class TxtPlotter(Plotter):
    """A plotter writing text-based output.

    This class will just write text-based output. It is similar to
    (:py:class:`.MplPlotter`) in the sense that the same functions are
    supported. Here, however, we do not plot any figures, we just
    write a column based text file with the results.

    Attributes
    ----------
    out_fmt : string
        Selects format for output plots.

    """

    def __init__(self, out_fmt, backup=False, out_dir=None):
        """Initialise the text writer.

        Parameters
        ----------
        out_fmt : string
            This is the format to use for the images.
        backup : boolean, optional
            Determines if we should overwrite or backup old files.
        out_dir : string, optional
            Determines if we should write the files to a particular
            directory.

        """
        super().__init__(backup=backup, plotter_type='text',
                         out_dir=out_dir)
        self.out_fmt = out_fmt

    def output_flux(self, results):
        """Store the output from the flux analysis in text files.

        Parameters
        ----------
        results : dict
            This is the dict with the results from the flux analysis.

        Returns
        -------
        outputfile : list of dicts
            A list containing the files created for the flux and for
            the error in the flux.

        """
        outfiles = []
        # make running average txt and error txt:
        for i in range(len(results['flux'])):
            flux = results['flux'][i]
            runflux = results['runflux'][i]
            errflux = results['errflux'][i]
            outfile = name_file(FLUXFILES['runflux'].format(i + 1),
                                self.out_fmt,
                                path=self.out_dir)
            outfiles.append(outfile)
            # output running average:
            txt_save_columns(outfile, 'Time, running average',
                             (flux[:, 0], runflux), backup=self.backup)
            # output block-error results:
            outfile = name_file(FLUXFILES['block'].format(i + 1),
                                self.out_fmt,
                                path=self.out_dir)
            outfiles.append(outfile)
            _txt_block_error(outfile, 'Block error for flux analysis',
                             errflux, backup=self.backup)
        return outfiles

    def output_energy(self, results, energies):
        """Save the output from the energy analysis to text files.

        Parameters
        ----------
        results : dict
            Each item in `results` contains the results for the
            corresponding energy. It is assumed to contains the keys
            'vpot', 'ekin', 'etot', 'ham', 'temp', 'elec'.
        energies : numpy.array
            This is the raw data for the energy analysis.

        Returns
        -------
        outfiles : list
            The output files created by this function.

        """
        outfiles = []
        time = energies['time']
        # 1) Store the running average:
        header = ['Running average of energy data: time']
        data = [time]
        for key in ['vpot', 'ekin', 'etot', 'ham', 'temp', 'ext']:
            if key in results:
                data.append(results[key]['running'])
                header.append(key)
        headertxt = ' '.join(header)
        outfile = name_file(ENERFILES['run_energies'], self.out_fmt,
                            path=self.out_dir)
        outfiles.append(outfile)
        txt_save_columns(outfile, headertxt, data, backup=self.backup)
        # 2) Save block error data:
        for key in ['vpot', 'ekin', 'etot', 'temp']:
            if key in results:
                outfile = name_file(ENERFILES['block'].format(key),
                                    self.out_fmt, path=self.out_dir)
                outfiles.append(outfile)
                _txt_block_error(outfile, ENERTITLE[key],
                                 results[key]['blockerror'],
                                 backup=self.backup)
        # 3) Save histograms:
        for key in ['vpot', 'ekin', 'etot', 'temp']:
            if key in results:
                outfile = name_file(ENERFILES['dist'].format(key),
                                    self.out_fmt, path=self.out_dir)
                outfiles.append(outfile)
                _txt_histogram(outfile,
                               f'Histogram for {ENERTITLE[key]}',
                               [results[key]['distribution']],
                               backup=self.backup)
        return outfiles

    def output_orderp(self, results, orderdata):
        """Save the output from the order parameter analysis to text files.

        Parameters
        ----------
        results : dict
            Each item in `results` contains the results for the
            corresponding order parameter.
        orderdata : numpy.array
            This is the raw data for the order parameter analysis

        Returns
        -------
        outfiles : list
            The output files created by this function.

        Note
        ----
        We are here only outputting results for the first order parameter.
        I.e. other order parameters or velocities are not written here. This
        will be changed when the structure of the output order parameter
        file has been fixed. Also note that, if present, the first order
        parameter will be plotted against the second one - i.e. the second
        one will be assumed to represent the velocity here.

        """
        outfiles = []
        time = orderdata[:, 0]
        # output running average:
        outfile = name_file(ORDERFILES['run_order'], self.out_fmt,
                            path=self.out_dir)
        txt_save_columns(outfile, 'Time, running average',
                         (time, results[0]['running']),
                         backup=self.backup)
        outfiles.append(outfile)

        # output block-error results:
        outfile = name_file(ORDERFILES['block'], self.out_fmt,
                            path=self.out_dir)
        _txt_block_error(outfile, 'Block error for order param',
                         results[0]['blockerror'], backup=self.backup)
        outfiles.append(outfile)
        # output distributions:
        outfile = name_file(ORDERFILES['dist'], self.out_fmt,
                            path=self.out_dir)
        _txt_histogram(outfile, 'Order parameter',
                       [results[0]['distribution']], backup=self.backup)
        outfiles.append(outfile)
        # output msd if it was calculated:
        if 'msd' in results[0]:
            msd = results[0]['msd']
            outfile = name_file(ORDERFILES['msd'], self.out_fmt,
                                path=self.out_dir)
            txt_save_columns(outfile, 'Time MSD Std',
                             (time[:len(msd)], msd[:, 0], msd[:, 1]),
                             backup=self.backup)
            outfiles.append(outfile)
            # TODO: time c/should here be multiplied with the correct dt
        return outfiles

    def output_path(self, results, path_ensemble):
        """Output all the results obtained by the path analysis.

        Parameters
        ----------
        results : dict
            This dict contains the result from the analysis.
        path_ensemble : object like :py:class:`.PathEnsemble`
            This is the path ensemble we have analysed.

        Returns
        -------
        outfiles : list
            The output files created by this function.

        """
        ens = path_ensemble.ensemble_name  # identify the ensemble
        ens_simplified = path_ensemble.ensemble_name_simple
        detect = results['detect']
        outfiles = []
        if 'pcross' in results:
            # 1) Output pcross vs lambda:
            outfile = name_file(PATHFILES['pcross'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            txt_save_columns(outfile,
                             f'Ensemble: {ens}, detect: {detect}',
                             [results['pcross'][0], results['pcross'][1]],
                             backup=self.backup)
        if 'prun' in results:
            # 2) Output the running average of p:
            outfile = name_file(PATHFILES['prun'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            txt_save_columns(outfile, f'Ensemble: {ens}',
                             [results['prun']], backup=self.backup)
        if 'blockerror' in results:
            # 3) Block error results:
            outfile = name_file(PATHFILES['perror'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            _txt_block_error(outfile, f'Ensemble: {ens}',
                             results['blockerror'], backup=self.backup)
        # 3) Length histograms
        outfile = name_file(PATHFILES['pathlength'].format(ens_simplified),
                            self.out_fmt, path=self.out_dir)
        outfiles.append(outfile)
        _txt_histogram(outfile, 'Histograms for acc and all',
                       [results['pathlength'][0], results['pathlength'][1]],
                       backup=self.backup)
        # 4) Shoot histograms
        outfile = name_file(PATHFILES['shoots'].format(ens_simplified),
                            self.out_fmt, path=self.out_dir)
        outfiles.append(outfile)
        _txt_shoots_histogram(outfile, results['shoots'][0],
                              results['shoots'][1], ens, backup=self.backup)
        return outfiles

    def output_pppath(self, results, path_ensemble):
        """Output all the results obtained by the repptis path analysis.

        Parameters
        ----------
        results : dict
            This dict contains the result from the analysis.
        path_ensemble : object like :py:class:`.PathEnsemble`
            This is the path ensemble we have analysed.

        Returns
        -------
        outfiles : list
            The output files created by this function.

        """
        ens = path_ensemble.ensemble_name  # identify the ensemble
        ens_simplified = path_ensemble.ensemble_name_simple
        detect = results['detect']
        outfiles = []
        if 'pcross' in results:
            # 1) Output pcross vs lambda:
            outfile = name_file(PATHFILES['pcross'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            txt_save_columns(outfile,
                             f'Ensemble: {ens}, detect: {detect}',
                             [results['pcross'][0], results['pcross'][1]],
                             backup=self.backup)
        if 'prun_sl' in results:
            # 2) Output the running average of p:
            outfile = name_file(PATHFILES['prun_sl'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            txt_save_columns(outfile, f'Ensemble: {ens}',
                             [results['prun_sl']], backup=self.backup)
        if 'prun_sr' in results:
            # 2) Output the running average of p:
            outfile = name_file(PATHFILES['prun_sr'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            txt_save_columns(outfile, f'Ensemble: {ens}',
                             [results['prun_sr']], backup=self.backup)
        if 'blockerror_sl' in results:
            # 3) Block error results:
            outfile = name_file(PATHFILES['perror_sl'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            _txt_block_error(outfile, f'Ensemble: {ens}',
                             results['blockerror_sl'], backup=self.backup)
        if 'blockerror_sr' in results:
            # 3) Block error results:
            outfile = name_file(PATHFILES['perror_sr'].format(ens_simplified),
                                self.out_fmt, path=self.out_dir)
            outfiles.append(outfile)
            _txt_block_error(outfile, f'Ensemble: {ens}',
                             results['blockerror_sr'], backup=self.backup)
        # 3) Length histograms
        outfile = name_file(PATHFILES['pathlength'].format(ens_simplified),
                            self.out_fmt, path=self.out_dir)
        outfiles.append(outfile)
        _txt_histogram(outfile, 'Histograms for acc and all',
                       [results['pathlength'][0], results['pathlength'][1]],
                       backup=self.backup)
        # 4) Shoot histograms
        outfile = name_file(PATHFILES['shoots'].format(ens_simplified),
                            self.out_fmt, path=self.out_dir)
        outfiles.append(outfile)
        _txt_shoots_histogram(outfile, results['shoots'][0],
                              results['shoots'][1], ens, backup=self.backup)
        return outfiles

    def output_matched_probability(self, path_ensembles, detect, matched,
                                   reptis=False):
        """Output the matched probabilities to a text file.

        This function will output the matched probabilities for the
        different ensembles and also output the over-all matched
        probability.

        Parameters
        ----------
        path_ensembles : list of strings
            This is the names for the path ensembles we have calculated
            the probability for.
        detect : list of floats
            These are the detect interfaces used in the analysis.
        matched : dict
            This dict contains the results from the matching of the
            probabilities. We make use of `matched['overall-prob']` and
            `matched['matched-prob']` here.

        Returns
        -------
        outfiles : list
            The files created by this function.

        """
        outfiles = []
        # start by creating the matched file, here we use a custom
        # file writer:
        if not reptis:
            outfile = name_file(PATH_MATCH['match'], self.out_fmt,
                                path=self.out_dir)
            use_gzip = outfile.endswith('.gz')
            if use_gzip:
                outfile = outfile[:-3]
            with open(outfile, 'wb') as fhandle:
                for prob, ens, idet in zip(matched['matched-prob'],
                                           path_ensembles, detect):
                    header = f'Ensemble: {ens}, idetect: {idet}'
                    np.savetxt(fhandle, prob, header=header)
            if use_gzip:
                with open(outfile, 'rb') as fhandle:
                    with gzip.open(f'{outfile}.gz', 'wb') as fhandle_out:
                        shutil.copyfileobj(fhandle, fhandle_out)
                os.remove(outfile)
                outfile = f'{outfile}.gz'
            outfiles.append(outfile)
        # output the over-all matched probability:
        if reptis:
            matched['overall-prob'] = np.array([detect,
                                                matched['pcross']]).T
        outfile = name_file(PATH_MATCH['total'], self.out_fmt,
                            path=self.out_dir)
        interf = ' , '.join([str(idet) for idet in detect])
        header = 'Total matched probability. Interfaces: {}'
        txt_save_columns(outfile, header.format(interf),
                         (matched['overall-prob'][:, 0],
                          matched['overall-prob'][:, 1]),
                         backup=self.backup)
        outfiles.append(outfile)

        # output the prun matched probability:
        if 'overall-rrun' in matched:
            outfile = name_file('overall-rrun', self.out_fmt,
                                path=self.out_dir)
            interf = ' , '.join([str(idet) for idet in detect])
            header = 'Running average overall rate. Interfaces: {}'
            txt_save_columns(outfile, header.format(interf),
                             (matched['overall-cycle'],
                              matched['overall-rrun']),
                             backup=self.backup)
            outfiles.append(outfile)
        if 'overall-prun' in matched:
            outfile = name_file('overall-prun', self.out_fmt,
                                path=self.out_dir)
            interf = ' , '.join([str(idet) for idet in detect])
            header = 'Running average crossing probability. Interfaces: {}'
            txt_save_columns(outfile, header.format(interf),
                             (matched['overall-cycle'],
                              matched['overall-prun']),
                             backup=self.backup)
            outfiles.append(outfile)

        return outfiles


def _txt_block_error(outputfile, title, error, backup=False):
    """Write the output from the error analysis to a text file.

    Parameters
    ----------
    outputfile : string
        This is the name of the output file to create.
    title : string
        This is an identifier/title to add to the header, e.g.
        'Ensemble: 001', 'Kinetic energy', etc.
    error : list
        This is the result of the error analysis.
    backup : boolean, optional
        Determines if we will do a backup of old files or not.

    """
    header = '{0}, Rel.err: {1:9.6e}, Ncor: {2:9.6f}'
    header = header.format(title, error[4], error[6])
    txt_save_columns(outputfile, header, (error[0], error[3]), backup=backup)


def _txt_histogram(outputfile, title, histograms, backup=False):
    """Write histograms to a text file.

    Parameters
    ----------
    outputfile : string
        This is the name of the output file to create.
    title : string
        A descriptive title to add to the header.
    histograms : tuple or list
        The histograms to store.
    backup : boolean, optional
        Determines if we will do a backup of old files or not.

    """
    data = []
    header = [f'{title}']
    for hist in histograms:
        header.append(f'avg: {hist[2][0]:6.2f}, std: {hist[2][1]:6.2f}')
        data.append(hist[1])
        data.append(hist[0])
    headertxt = ', '.join(header)
    txt_save_columns(outputfile, headertxt, data, backup=backup)


def _txt_shoots_histogram(outputfile, histograms, scale, ensemble,
                          backup=False):
    """Write the histograms from the shoots analysis to a text file.

    Parameters
    ----------
    histograms : dict
        These are the histograms obtained in the shoots analysis.
    scale : dict
        These are the scale factors for normalising the histograms
        obtained in the shoots analysis.
    ensemble : string
        This is the ensemble identifier, e.g. 001, 002, etc.
    outputfile : string
        This is the name of the output file to create.
    backup : boolean, optional
        Determines if we will do a backup of old files or not.

    """
    data = []
    header = [f'Ensemble: {ensemble}']
    for key in ['ACC', 'REJ', 'BWI', 'ALL']:
        try:
            mid = histograms[key][2]
            hist = histograms[key][0]
            hist_scale = hist * scale[key]
            data.append(mid)
            data.append(hist)
            data.append(hist_scale)
            header.append(f'{key} (mid, hist, hist*scale)')
        except KeyError:
            continue
    headertxt = ', '.join(header)
    txt_save_columns(outputfile, headertxt, data, backup=backup)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A GROMACS external MD integrator interface.

This module defines a class for using GROMACS as an external engine.

Important classes defined here
------------------------------

GromacsEngine (:py:class:`.GromacsEngine`)
    A class responsible for interfacing GROMACS.
"""
import logging
import os
import shlex
from pyretis.core.box import box_matrix_to_list
from pyretis.engines.external import ExternalMDEngine
from pyretis.inout.settings import look_for_input_files
from pyretis.inout.formats.gromacs import (
    read_gromos96_file,
    read_gromacs_gro_file,
    write_gromacs_gro_file,
    write_gromos96_file,
    read_xvg_file,
    read_trr_frame,
)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class GromacsEngine(ExternalMDEngine):
    """
    A class for interfacing GROMACS.

    This class defines the interface to GROMACS.

    Attributes
    ----------
    gmx : string
        The command for executing GROMACS. Note that we are assuming
        that we are using version 5 (or later) of GROMACS.
    mdrun : string
        The command for executing GROMACS mdrun. In some cases, this
        executable can be different from ``gmx mdrun``.
    mdrun_c : string
        The command for executing GROMACS mdrun when continuing a
        simulation. This is derived from the ``mdrun`` command.
    input_path : string
        The directory where the input files are stored.
    subcycles : int,
        The number of simulation steps of the external engine for each
        PyRETIS step (e.g. interaction between the softwares frequency)
    exe_path : string, optional
        The absolute path at which the main PyRETIS simulation will be run.
    maxwarn : integer
        Setting for the GROMACS ``grompp -maxwarn`` option.
    gmx_format : string
        This string selects the output format for GROMACS.
    write_vel : boolean, optional
        True if we want to output the velocities.
    write_force : boolean, optional
        True if we want to output the forces.
    domain_decomp: boolean, optional
        Whether gromacs uses domain decomposition (DD) or not. This
        does not set domain decomposition. The user should know whether
        DD is executed or not.
    """

    def __init__(self, gmx, mdrun, input_path, timestep, subcycles,
                 exe_path=os.path.abspath('.'),
                 maxwarn=0, gmx_format='gro',
                 write_vel=True,
                 write_force=False,
                 domain_decomp=False):
        """Set up the GROMACS engine.

        Parameters
        ----------
        gmx : string
            The GROMACS executable.
        mdrun : string
            The GROMACS mdrun executable.
        input_path : string
            The absolute path to where the input files are stored.
        timestep : float
            The time step used in the GROMACS MD simulation.
        subcycles : integer
            The number of steps each GROMACS MD run is composed of.
        exe_path : string, optional
            The absolute path at which the main PyRETIS simulation will be run.
        maxwarn : integer, optional
            Setting for the GROMACS ``grompp -maxwarn`` option.
        gmx_format : string, optional
            The format used for GROMACS configurations.
        write_vel : boolean, optional
            Determines if GROMACS should write velocities or not.
        write_force : boolean, optional
            Determines if GROMACS should write forces or not.
        domain_decomp: boolean, optional
            Whether domain decomposition (DD) is executed by GROMACS or not.
            User should know if this happens or not.
        """
        super().__init__('GROMACS engine', timestep, subcycles)
        self.ext = gmx_format
        if self.ext not in ('g96', 'gro'):
            msg = 'Unknown GROMACS format: "%s"'
            logger.error(msg, self.ext)
            raise ValueError(msg % self.ext)
        # Define the GROMACS GMX command:
        self.gmx = gmx
        # Define GROMACS GMX MDRUN commands:
        self.mdrun = mdrun + ' -s {} -deffnm {} -c {}'
        # This is for continuation of a GROMACS simulation:
        self.mdrun_c = mdrun + ' -s {} -cpi {} -append -deffnm {} -c {}'
        # Whether DD is used or not:
        self.domain_decomp = domain_decomp
        self.ext_time = self.timestep * self.subcycles
        self.maxwarn = maxwarn
        # Define the energy terms, these are hard-coded, but
        # here we open up for changing that:
        self.energy_terms = self.select_energy_terms('path')
        self.input_path = os.path.join(exe_path, input_path)
        # Set the defaults input files:
        default_files = {
            'conf': f'conf.{self.ext}',
            'input_o': 'grompp.mdp',  # "o" = original input file.
            'topology': 'topol.top'}
        extra_files = {
            'index': 'index.ndx',
        }

        # An user doesn't need to have problems with g96 and mdtraj.
        file_g = os.path.join(self.input_path, 'conf.')
        if self.ext == 'gro':
            self.top, _, _, _ = read_gromacs_gro_file(file_g+self.ext)
        elif self.ext == 'g96':
            if not os.path.isfile(file_g+'gro'):
                cmd = [self.gmx, 'editconf',
                       '-f', file_g+self.ext,
                       '-o', file_g+'gro']
                self.execute_command(cmd, cwd=None)

            self.top, _, _, _ = read_gromos96_file(file_g+self.ext)
            self.top['VELOCITY'] = self.top['POSITION'].copy()

        # Check the presence of the defaults input files or, if absent,
        # try to find them by their expected extension.
        self.input_files = look_for_input_files(self.input_path,
                                                default_files,
                                                extra_files)
        logger.debug('Input files: %s', self.input_files)
        # Check the input file and create a PyRETIS version with
        # consistent settings:
        settings = {
            'dt': self.timestep,
            'nstxout-compressed': 0,
            'gen_vel': 'no'
        }
        for key in ('nsteps', 'nstxout', 'nstvout', 'nstfout', 'nstlog',
                    'nstcalcenergy', 'nstenergy'):
            settings[key] = self.subcycles
        if not write_vel:
            settings['nstvout'] = 0
        if not write_force:
            settings['nstfout'] = 0

        # PyRETIS construct its own mdp file
        self.input_files['input'] = os.path.join(self.input_path,
                                                 'pyretis.mdp')
        self._modify_input(self.input_files['input_o'],
                           self.input_files['input'], settings, delim='=')
        logger.info(('Created GROMACS mdp input from %s. You might '
                     'want to check the input file: %s'),
                    self.input_files['input_o'], self.input_files['input'])

        # Generate a tpr file using the input files:
        logger.info('Creating ".tpr" for GROMACS in %s', self.input_path)
        self.exe_dir = self.input_path
        out_files = self._execute_grompp(self.input_files['input'],
                                         self.input_files['conf'], 'topol')

        # This will generate some noise, let's remove files we don't need:
        mdout = os.path.join(self.input_path, out_files['mdout'])
        self._removefile(mdout)
        # We also remove GROMACS backup files after creating the tpr:
        self._remove_gromacs_backup_files(self.input_path)
        # Keep the tpr file.
        self.input_files['tpr'] = os.path.join(self.input_path,
                                               out_files['tpr'])
        logger.info('GROMACS ".tpr" created: %s', self.input_files['tpr'])

    @staticmethod
    def select_energy_terms(terms):
        """Select energy terms to extract from GROMACS.

        Parameters
        ----------
        terms : string
            This string will name the terms to extract. Currently
            we only allow for two types of output, but this can be
            customized in the future.

        """
        allowed_terms = {
            'full': ('\n'.join(('Potential', 'Kinetic-En.', 'Total-Energy',
                                'Temperature', 'Pressure'))).encode(),
            'path': b'Potential\nKinetic-En.',
        }
        if terms not in allowed_terms:
            return allowed_terms['path']
        return allowed_terms[terms]

    @staticmethod
    def rename_energies(gmx_energy):
        """Rename GROMACS energy terms to PyRETIS convention."""
        energy_map = {'potential': 'vpot',
                      'kinetic en.': 'ekin',
                      'temperature': 'temp',
                      'total energy': 'etot',
                      'pressure': 'press'}
        energy = {}
        for key, val in gmx_energy.items():
            name = energy_map.get(key, key)
            energy[name] = val[0]
        return energy

    def _execute_grompp(self, mdp_file, config, deffnm):
        """Execute the GROMACS preprocessor.

        Parameters
        ----------
        mdp_file : string
            The path to the mdp file.
        config : string
            The path to the GROMACS config file to use as input.
        deffnm : string
            A string used to name the GROMACS files.

        Returns
        -------
        out_files : dict
            This dict contains files that were created by the GROMACS
            preprocessor.

        """
        topol = self.input_files['topology']
        tpr = f'{deffnm}.tpr'
        cmd = [self.gmx, 'grompp', '-f', mdp_file, '-c', config,
               '-p', topol, '-o', tpr]
        if 'index' in self.input_files:
            cmd.extend(['-n', self.input_files['index']])
        if self.maxwarn > 0:
            cmd.extend(['-maxwarn', str(self.maxwarn)])
        self.execute_command(cmd, cwd=self.exe_dir)
        out_files = {'tpr': tpr, 'mdout': 'mdout.mdp'}
        return out_files

    def _execute_mdrun(self, tprfile, deffnm):
        """
        Execute GROMACS mdrun.

        This method is intended as the initial ``gmx mdrun`` executed.
        That is, we here assume that we do not continue a simulation.

        Parameters
        ----------
        tprfile : string
            The .tpr file to use for executing GROMACS.
        deffnm : string
            To give the GROMACS simulation a name.

        Returns
        -------
        out_files : dict
            This dict contains the output files created by ``mdrun``.
            Note that we here hard code the file names.

        """
        confout = f'{deffnm}.{self.ext}'
        cmd = shlex.split(self.mdrun.format(tprfile, deffnm, confout))
        self.execute_command(cmd, cwd=self.exe_dir)
        out_files = {'conf': confout,
                     'cpt_prev': f'{deffnm}_prev.cpt'}
        for key in ('cpt', 'edr', 'log', 'trr'):
            out_files[key] = f'{deffnm}.{key}'
        self._remove_gromacs_backup_files(self.exe_dir)
        return out_files

    def _execute_grompp_and_mdrun(self, config, deffnm):
        """
        Execute GROMACS ``grompp`` and ``mdrun``.

        Here we use the input file given in the input directory.

        Parameters
        ----------
        config : string
            The path to the GROMACS config file to use as input.
        deffnm : string
            A string used to name the GROMACS files.

        Returns
        -------
        out_files : dict of strings
            The files created by this command.

        """
        out_files = {}
        out_grompp = self._execute_grompp(self.input_files['input'],
                                          config, deffnm)
        tpr_file = out_grompp['tpr']
        for key, value in out_grompp.items():
            out_files[key] = value
        out_mdrun = self._execute_mdrun(tpr_file,
                                        deffnm)
        for key, value in out_mdrun.items():
            out_files[key] = value
        return out_files

    def _execute_mdrun_continue(self, tprfile, cptfile, deffnm):
        """
        Continue the execution of GROMACS.

        Here, we assume that we have already executed ``gmx mdrun`` and
        that we are to append and continue a simulation.

        Parameters
        ----------
        tprfile : string
            The .tpr file which defines the simulation.
        cptfile : string
            The last checkpoint file (.cpt) from the previous run.
        deffnm : string
            To give the GROMACS simulation a name.

        Returns
        -------
        out_files : dict
            The output files created/appended by GROMACS when we
            continue the simulation.

        """
        confout = f'{deffnm}.{self.ext}'.format(deffnm, self.ext)
        self._removefile(confout)
        cmd = shlex.split(self.mdrun_c.format(tprfile, cptfile,
                                              deffnm, confout))
        self.execute_command(cmd, cwd=self.exe_dir)
        out_files = {'conf': confout}
        for key in ('cpt', 'edr', 'log', 'trr'):
            out_files[key] = f'{deffnm}.{key}'
        self._remove_gromacs_backup_files(self.exe_dir)
        return out_files

    def _extend_gromacs(self, tprfile, time):
        """Extend a GROMACS simulation.

        Parameters
        ----------
        tprfile : string
            The file to read for extending.
        time : float
            The time (in ps) to extend the simulation by.

        Returns
        -------
        out_files : dict
            The files created by GROMACS when we extend.

        """
        tpxout = f'ext_{tprfile}'
        self._removefile(tpxout)
        cmd = [self.gmx, 'convert-tpr', '-s', tprfile,
               '-extend', str(time), '-o', tpxout]
        self.execute_command(cmd, cwd=self.exe_dir)
        out_files = {'tpr': tpxout}
        return out_files

    def _extend_and_execute_mdrun(self, tpr_file, cpt_file, deffnm):
        """Extend GROMACS and execute mdrun.

        Parameters
        ----------
        tpr_file : string
            The location of the "current" .tpr file.
        cpt_file : string
            The last checkpoint file (.cpt) from the previous run.
        deffnm : string
            To give the GROMACS simulation a name.

        Returns
        -------
        out_files : dict
            The files created by GROMACS when we extend.

        """
        out_files = {}
        out_grompp = self._extend_gromacs(tpr_file, self.ext_time)
        ext_tpr_file = out_grompp['tpr']
        for key, value in out_grompp.items():
            out_files[key] = value
        out_mdrun = self._execute_mdrun_continue(ext_tpr_file, cpt_file,
                                                 deffnm)
        for key, value in out_mdrun.items():
            out_files[key] = value
        # Move extended tpr so that we can continue extending:
        source = os.path.join(self.exe_dir, ext_tpr_file)
        dest = os.path.join(self.exe_dir, tpr_file)
        self._movefile(source, dest)
        out_files['tpr'] = tpr_file
        return out_files

    def _remove_gromacs_backup_files(self, dirname):
        """Remove files GROMACS has backed up.

        These are files starting with a '#'

        Parameters
        ----------
        dirname : string
            The directory where we are to remove files.

        """
        for entry in os.scandir(dirname):
            if entry.name.startswith('#') and entry.is_file():
                filename = os.path.join(dirname, entry.name)
                self._removefile(filename)

    def _extract_frame(self, traj_file, idx, out_file):
        """Extract a frame from a .trr, .xtc or .trj file.

        If the extension is different from .trr, .xtc or .trj, we will
        basically just copy the given input file.

        Parameters
        ----------
        traj_file : string
            The GROMACS file to open.
        idx : integer
            The frame number we look for.
        out_file : string
            The file to dump to.

        Note
        ----
        This will only properly work if the frames in the input
        trajectory are uniformly spaced in time.

        """
        trajexts = ['.trr', '.xtc', '.trj']

        logger.debug('Extracting frame, idx = %i', idx)
        logger.debug('Source file: %s, out file: %s', traj_file, out_file)
        if traj_file[-4:] in trajexts:
            _, data = read_trr_frame(traj_file, idx)
            xyz = data['x']
            vel = data.get('v')
            box = box_matrix_to_list(data['box'], full=True)
            if out_file[-4:] == '.gro':
                write_gromacs_gro_file(out_file, self.top, xyz, vel, box)
            elif out_file[-4:] == '.g96':
                write_gromos96_file(out_file, self.top, xyz, vel, box)

        else:
            cmd = [self.gmx, 'trjconv',
                   '-f', traj_file,
                   '-s', self.input_files['tpr'],
                   '-o', out_file]

            self.execute_command(cmd, inputs=b'0', cwd=None)

        # If domain decomposition is used, the molecules must be made
        # whole. This is a bug in GROMACS, where DD only works auto-
        # matically for initial/ending .gro files of a simulation
        # trajectory. I.e. only for the latter .gro 'frames', the bonds
        # are correctly represented over the periodic boundaries.
        # This is version specific and *could* happen only for >16cpus
        # Therefore, we don't test this in coverage...
        if self.domain_decomp:  # pragma: no cover
            cmd = [self.gmx, 'trjconv',
                   '-f', out_file,
                   '-s', self.input_files['tpr'],
                   '-pbc', 'whole',
                   '-o', out_file]
            logger.info(("We are making the molecules whole"))
            self.execute_command(cmd, inputs=b'0', cwd=None)

    def get_energies(self, energy_file, begin=None, end=None):
        """Return energies from a GROMACS run.

        Parameters
        ----------
        energy_file : string
            The file to read energies from.
        begin : float, optional
            Select the time for the first frame to read.
        end : float, optional
            Select the time for the last frame to read.

        Returns
        -------
        energy : dict fo numpy.arrays
            The energies read from the produced GROMACS xvg file.

        """
        cmd = [self.gmx, 'energy', '-f', energy_file]
        if begin is not None:
            begin = max(begin, 0)
            cmd.extend(['-b', str(begin)])
        if end is not None:
            cmd.extend(['-e', str(end)])
        self.execute_command(cmd, inputs=self.energy_terms,
                             cwd=self.exe_dir)
        xvg_file = os.path.join(self.exe_dir, 'energy.xvg')
        energy = read_xvg_file(xvg_file)
        self._removefile(xvg_file)
        return energy

    def _propagate_from(self, name, path, ensemble, msg_file, reverse=False):
        """
        Propagate with GROMACS from the current system configuration.

        Here, we assume that this method is called after the propagate()
        has been called in the parent. The parent is then responsible
        for reversing the velocities and also for setting the initial
        state of the system.

        Parameters
        ----------
        name : string
            A name to use for the trajectory we are generating.
        path : object like :py:class:`pyretis.core.path.PathBase`
            This is the path we use to fill in phase-space points.
        ensemble: dict
            It contains:

            * `system`: object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function`: object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `interfaces`: list of floats
              These interfaces define the stopping criterion.

        msg_file : object like :py:class:`.FileIO`
            An object we use for writing out messages that are useful
            for inspecting the status of the current propagation.
        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.

        """
        status = f'propagating with GROMACS (reverse = {reverse})'
        logger.debug(status)
        success = False
        interfaces = ensemble['interfaces']
        left, _, right = interfaces
        # Dumping of the initial config were done by the parent, here
        # we will just use it:
        initial_conf = ensemble['system'].particles.get_pos()[0]
        # Get the current order parameter:
        order = self.calculate_order(ensemble)
        msg_file.write(
            f'# Initial order parameter: {" ".join([str(i) for i in order])}'
        )
        # In some cases, we don't really have to perform a step as the
        # initial config might be left/right of the interface in
        # question. Here, we will perform a step anyway. This is to be
        # sure that we obtain energies and also a trajectory segment.
        # Note that all the energies are obtained after we are done
        # with the integration from the .edr file of the trajectory.
        msg_file.write('# Running grompp and mdrun (initial step).')
        out_files = self._execute_grompp_and_mdrun(initial_conf, name)
        # Define name of some files:
        tpr_file = out_files['tpr']
        cpt_file = out_files['cpt']
        traj_file = os.path.join(self.exe_dir, out_files['trr'])
        msg_file.write(f'# Trajectory file is: {traj_file}')
        conf_abs = os.path.join(self.exe_dir, out_files['conf'])
        # Note: The order parameter is calculated AT THE END of each iteration.
        msg_file.write('# Starting GROMACS.')
        msg_file.write('# Step order parameter cv1 cv2 ...')
        for i in range(path.maxlen):
            msg_file.write(
                f'{i} {" ".join([str(j) for j in order])}'
            )
            # We first add the previous phase point, and then we propagate.
            snapshot = {'order': order,
                        'config': (traj_file, i),
                        'vel_rev': reverse}
            phase_point = self.snapshot_to_system(ensemble['system'],
                                                  snapshot)
            status, success, stop, _ = self.add_to_path(path, phase_point,
                                                        left, right)
            if stop:
                logger.debug('GROMACS propagation ended at %i. Reason: %s',
                             i, status)
                break
            if i == 0:
                # This step was performed before entering the main loop.
                pass
            elif i > 0:
                out_extnd = self._extend_and_execute_mdrun(tpr_file, cpt_file,
                                                           name)
                out_files.update(out_extnd)
            # Calculate the order parameter using the current system:
            ensemble['system'].particles.set_vel(reverse)
            ensemble['system'].particles.set_pos((conf_abs, None))
            order = self.calculate_order(ensemble)
            # We now have the order parameter, for GROMACS just remove the
            # config file to avoid the GROMACS #conf_abs# backup clutter:
            self._removefile(conf_abs)
            msg_file.flush()
        logger.debug('GROMACS propagation done, obtaining energies')
        msg_file.write('# Propagation done.')
        msg_file.write(f'# Reading energies from: {out_files["edr"]}')
        msg_file.flush()
        energy = self.get_energies(out_files['edr'])
        path.update_energies(energy['kinetic en.'], energy['potential'])
        logger.debug('Removing GROMACS output after propagate.')
        remove = [val for key, val in out_files.items()
                  if key not in ('trr', 'gro', 'g96')]
        self._remove_files(self.exe_dir, remove)
        self._remove_gromacs_backup_files(self.exe_dir)
        return success, status

    def step(self, system, name):
        """Perform a single step with GROMACS.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we are integrating.
        name : string
            To name the output files from the GROMACS step.

        Returns
        -------
        out : string
            The name of the output configuration, obtained after
            completing the step.

        """
        initial_conf = self.dump_frame(system)
        # Save as a single snapshot file:
        system.particles.set_pos((initial_conf, None))
        system.particles.set_vel(False)
        out_grompp = self._execute_grompp(self.input_files['input'],
                                          initial_conf,
                                          name)
        out_mdrun = self._execute_mdrun(out_grompp['tpr'], name)
        conf_abs = os.path.join(self.exe_dir, out_mdrun['conf'])
        logger.debug('Obtaining GROMACS energies after single step.')
        energy = self.get_energies(out_mdrun['edr'])
        system.particles.set_pos((conf_abs, None))
        system.particles.set_vel(False)
        system.particles.vpot = energy['potential'][-1]
        system.particles.ekin = energy['kinetic en.'][-1]
        logger.debug('Removing GROMACS output after single step.')
        remove = [val for _, val in out_grompp.items()]
        remove += [val for key, val in out_mdrun.items() if key != 'conf']
        self._remove_files(self.exe_dir, remove)
        return out_mdrun['conf']

    def _prepare_shooting_point(self, input_file):
        """
        Create the initial configuration for a shooting move.

        This creates a new initial configuration with random velocities.
        Here, the random velocities are obtained by running a zero-step
        GROMACS simulation.

        Parameters
        ----------
        input_file : string
            The input configuration to generate velocities for.

        Returns
        -------
        output_file : string
            The name of the file created.
        energy : dict
            The energy terms read from the GROMACS .edr file.

        """
        gen_mdp = os.path.join(self.exe_dir, 'genvel.mdp')
        if os.path.isfile(gen_mdp):
            logger.debug('%s found. Re-using it!', gen_mdp)
        else:
            # Create output file to generate velocities:
            settings = {'gen_vel': 'yes', 'gen_seed': -1, 'nsteps': 0,
                        'continuation': 'no'}
            self._modify_input(self.input_files['input'], gen_mdp, settings,
                               delim='=')
        # Run GROMACS grompp for this input file:
        out_grompp = self._execute_grompp(gen_mdp, input_file, 'genvel')
        remove = [val for _, val in out_grompp.items()]
        # Run GROMACS mdrun for this tpr file:
        out_mdrun = self._execute_mdrun(out_grompp['tpr'], 'genvel')
        remove += [val for key, val in out_mdrun.items() if key != 'conf']
        confout = os.path.join(self.exe_dir, out_mdrun['conf'])
        energy = self.get_energies(out_mdrun['edr'])
        # Remove run-files:
        logger.debug('Removing GROMACS output after velocity generation.')
        self._remove_files(self.exe_dir, remove)
        return confout, energy

    def _read_configuration(self, filename):
        """Read output from GROMACS .g96/gro files.

        Parameters
        ----------
        filename : string
            The file to read the configuration from.

        Returns
        -------
        box : numpy.array
            The box dimensions.
        xyz : numpy.array
            The positions.
        vel : numpy.array
            The velocities.

        """
        box = None
        if self.ext == 'g96':
            _, xyz, vel, box = read_gromos96_file(filename)
        elif self.ext == 'gro':
            _, xyz, vel, box = read_gromacs_gro_file(filename)
        else:
            msg = 'GROMACS engine does not support reading "%s"'
            logger.error(msg, self.ext)
            raise ValueError(msg % self.ext)
        return box, xyz, vel

    def _reverse_velocities(self, filename, outfile):
        """Reverse velocity in a given snapshot.

        Parameters
        ----------
        filename : string
            The configuration to reverse velocities in.
        outfile : string
            The output file for storing the configuration with
            reversed velocities.

        """
        if self.ext == 'g96':
            txt, xyz, vel, _ = read_gromos96_file(filename)
            write_gromos96_file(outfile, txt, xyz, -1 * vel)
        elif self.ext == 'gro':
            txt, xyz, vel, _ = read_gromacs_gro_file(filename)
            write_gromacs_gro_file(outfile, txt, xyz, -1 * vel)
        else:
            msg = 'GROMACS engine does not support writing "%s"'
            logger.error(msg, self.ext)
            raise ValueError(msg % self.ext)

    def modify_velocities(self, ensemble, vel_settings):
        """Modify the velocities of the current state.

        This method will modify the velocities of a time slice.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system`: object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating.

        vel_settings: dict
            It contains:

            * `sigma_v`: numpy.array, optional
              These values can be used to set a standard deviation (one
              for each particle) for the generated velocities.
            * `aimless`: boolean, optional
              Determines if we should do aimless shooting or not.
            * `momentum`: boolean, optional
              If True, we reset the linear momentum to zero after
              generating.
            * `rescale or rescale_energy`: float, optional
              In some NVE simulations, we may wish to re-scale the
              energy to a fixed value. If `rescale` is a float > 0,
              we will re-scale the energy (after modification of
              the velocities) to match the given float.

        Returns
        -------
        dek : float
            The change in the kinetic energy.
        kin_new : float
            The new kinetic energy.

        """
        dek = None
        kin_old = None
        kin_new = None
        system = ensemble['system']
        rescale = vel_settings.get('rescale_energy',
                                   vel_settings.get('rescale'))
        if rescale is not None and rescale is not False and rescale > 0:
            msgtxt = 'GROMACS engine does not support energy re-scale.'
            logger.error(msgtxt)
            raise NotImplementedError(msgtxt)
        kin_old = system.particles.ekin
        if vel_settings.get('aimless', False):
            pos = self.dump_frame(system)
            posvel, energy = self._prepare_shooting_point(pos)
            kin_new = energy['kinetic en.'][-1]
            system.particles.set_pos((posvel, None))
            system.particles.set_vel(False)
            system.particles.ekin = kin_new
            system.particles.vpot = energy['potential'][-1]
        else:  # Soft velocity change, from a Gaussian distribution:
            msgtxt = 'GROMACS engine only support aimless shooting!'
            logger.error(msgtxt)
            raise NotImplementedError(msgtxt)
        if vel_settings.get('momentum', False):
            pass
        if kin_old is None or kin_new is None:
            dek = float('inf')
            logger.debug(('Kinetic energy not found for previous point.'
                          '\n(This happens when the initial configuration '
                          'does not contain energies.)'))
        else:
            dek = kin_new - kin_old
        return dek, kin_new

    def integrate(self, ensemble, steps, thermo='full'):
        """
        Perform several integration steps.

        This method will perform several integration steps using
        GROMACS. It will also calculate order parameter(s) and energy
        terms if requested.

        Parameters
        ----------
        ensemble: dict
            It contains:

            * `system`: object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function`: object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.

        steps : integer
            The number of steps we are going to perform. Note that we
            do not integrate on the first step (e.g. step 0) but we do
            obtain the other properties. This is to output the starting
            configuration.
        thermo : string, optional
            Select the thermodynamic properties we are to calculate.

        Yields
        ------
        results : dict
            The results from a MD step. This contains the state of the system
            and order parameter(s) and energies (if calculated).

        """
        logger.debug('Integrating with GROMACS')
        # Dump the initial config:
        system = ensemble['system']
        order_function = ensemble.get('order_function')
        initial_file = self.dump_frame(system)
        out_files = {}
        conf_abs = None
        self.energy_terms = self.select_energy_terms(thermo)
        # For step zero, obtain the order parameter:
        if order_function:
            order = self.calculate_order(ensemble)
        else:
            order = None

        for i in range(steps):
            if i == 0:
                out_files = self._execute_grompp_and_mdrun(
                    initial_file,
                    'pyretis-gmx'
                )
                conf_abs = os.path.join(self.exe_dir, out_files['conf'])
            elif 0 < i < steps - 1:
                out_extnd = self._extend_and_execute_mdrun(
                    out_files['tpr'],
                    out_files['cpt'],
                    'pyretis-gmx'
                )
                out_files.update(out_extnd)
            else:
                pass
            # Update with results from previous step:
            results = {}
            if order:
                results['order'] = order
            # Update for order parameter:
            if order_function:
                system.particles.set_pos((conf_abs, None, None))
                order = self.calculate_order(ensemble)
            # Obtain latest energies:
            time1 = i * self.timestep * self.subcycles
            time2 = (i + 1) * self.timestep * self.subcycles
            # time1 and time2 should be correct now, but we are victims
            # of floating points. Subtract/add something small so that
            # we round to correct time.
            time1 -= self.timestep * 0.1
            time2 += self.timestep * 0.1
            energy = self.get_energies(out_files['edr'], begin=time1,
                                       end=time2)
            # Rename energies into the PyRETIS convention:
            results['thermo'] = self.rename_energies(energy)
            yield results



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A script for building the C extension (order parameter and potential)."""
from setuptools import (
    setup,
    Extension,
)
import numpy as np


WCAORDER = Extension(
    'wcalambda',
    sources=['wcalambda.c'],
    extra_compile_args=['-Ofast', '-march=native'],
)
WCAFORCE = Extension(
    'wcaforces',
    sources=['wcaforces.c'],
    extra_compile_args=['-Ofast', '-march=native'],
)
setup(
    name='PyRETIS WCA C extension',
    description='C extensions for the WCA potential.',
    ext_modules=[WCAORDER, WCAFORCE],
    include_dirs=[np.get_include()],
)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of engines.

This package defines engines for PyRETIS. The engines are responsible
for carrying out dynamics for a system. This can in principle both
be molecular dynamics or Monte Carlo dynamics. Typically, with RETIS,
this will be molecular dynamics in some form in order to propagate
the equations of motion and obtain new trajectories.


Package structure
-----------------

Modules
~~~~~~~

cp2k.py (:py:mod:`pyretis.engines.cp2k`)
    Defines an engine for use with CP2K.

engine.py (:py:mod:`pyretis.engines.engine`)
    Defines the base engine class.

external.py (:py:mod:`pyretis.engines.external`)
    Defines the interface for external engines.

gromacs.py (:py:mod:`pyretis.engines.gromacs`)
    Defines an engine for use with GROMACS.

gromacs2.py (:py:mod:`pyretis.engines.gromacs2`)
    Defines an engine for use with GROMACS. This is
    an alternative implementation which does not rely on
    continuously starting and stopping the GROMACS
    executable.

internal.py (:py:mod:`pyretis.engines.internal`)
    Defines internal PyRETIS engines.

lammps.py (:py:mod:`pyretis.engines.lammps`)
    Defines and engine for use with LAMMPS.

openmm.py (:py:mod:`pyretis.engines.openmm`)
    Defines an engine for use with OpenMM.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

engine_factory (:py:func:`.engine_factory`)
    A method to create engines from settings.
"""
from pyretis.core.common import generic_factory
from .internal import MDEngine, Verlet, VelocityVerlet, Langevin
from .external import ExternalMDEngine
from .gromacs import GromacsEngine
from .gromacs2 import GromacsEngine2
from .cp2k import CP2KEngine
from .openmm import OpenMMEngine
from .lammps import LAMMPSEngine


def engine_factory(settings):
    """Create an engine according to the given settings.

    This function is included as a convenient way of setting up and
    selecting an engine. It will return the created engine.

    Parameters
    ----------
    settings : dict
        This defines how we set up and select the engine.

    Returns
    -------
    out : object like :py:class:`.EngineBase`
        The object representing the engine to use in a simulation.

    """
    engine_map = {
        'velocityverlet': {'cls': VelocityVerlet},
        'verlet': {'cls': Verlet},
        'langevin': {'cls': Langevin},
        'gromacs': {'cls': GromacsEngine},
        'gromacs2': {'cls': GromacsEngine2},
        'cp2k': {'cls': CP2KEngine},
        'openmm': {'cls': OpenMMEngine},
        'lammps': {'cls': LAMMPSEngine},
    }
    return generic_factory(settings, engine_map, name='engine')



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definitions of simulation objects for path sampling simulations.

This module defines simulations for performing path sampling
simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

PathSimulation (:py:class:`.PathSimulation`)
    The base class for path simulations.

SimulationTIS (:py:class:`.SimulationSingleTIS`)
    Definition of a TIS simulation for a single path ensemble.

SimulationRETIS (:py:class:`.SimulationRETIS`)
    Definition of a RETIS simulation.

SimulationPPRETIS (:py:class:`.SimulationPPRETIS`)
    Definition of a PPRETIS simulation.

"""
import logging
import os
import numpy as np
from pyretis.core.common import soft_partial_exit, priority_checker
from pyretis.core.random_gen import create_random_generator
from pyretis.core.retis import make_retis_step
from pyretis.core.tis import make_tis
from pyretis.initiation import initiate_path_simulation
from pyretis.inout.common import make_dirs
from pyretis.inout.restart import write_ensemble_restart
from pyretis.inout.screen import print_to_screen
from pyretis.inout.simulationio import task_from_settings
from pyretis.simulation.simulation import Simulation

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['PathSimulation', 'SimulationTIS', 'SimulationRETIS']


class PathSimulation(Simulation):
    """A base class for TIS/RETIS simulations.

    Attributes
    ----------
    ensembles : list of dictionaries of objects
        Each contains:

        * `path_ensemble`: objects like :py:class:`.PathEnsemble`
          This is used for storing results for the different path ensembles.
        * `engine`: object like :py:class:`.EngineBase`
          This is the integrator that is used to propagate the system in time.
        * `rgen`: object like :py:class:`.RandomGenerator`
          This is a random generator used for the generation of paths.
        * `system`: object like :py:class:`.System`
          This is the system the simulation will act on.

    settings : dict
        A dictionary with TIS and RETIS settings. We expect that
        we can find ``settings['tis']`` and possibly
        ``settings['retis']``. For ``settings['tis']`` we further
        expect to find the keys:

        * `aimless`: Determines if we should do aimless shooting
          (True) or not (False).
        * `sigma_v`: Scale used for non-aimless shooting.
        * `seed`: A integer seed for the random generator used for
          the path ensemble we are simulating here.

        Note that the
        :py:func:`pyretis.core.tis.make_tis_step_ensemble` method
        will make use of additional keys from ``settings['tis']``.
        Please see this method for further details. For the
        ``settings['retis']`` we expect to find the following keys:

        * `swapfreq`: The frequency for swapping moves.
        * `relative_shoots`: If we should do relative shooting for
          the path ensembles.
        * `nullmoves`: Should we perform null moves.
        * `swapsimul`: Should we just swap a single pair or several pairs.

      required_settings : tuple of strings
        This is just a list of the settings that the simulation
        requires. Here it is used as a check to see that we have
        all we need to set up the simulation.

    """

    required_settings = ('tis', 'retis')
    name = 'Generic path simulation'
    simulation_type = 'generic-path'
    simulation_output = [
        {
            'type': 'pathensemble',
            'name': 'path_ensemble',
            'result': ('pathensemble-{}',),
        },
        {
            'type': 'path-order',
            'name': 'path_ensemble-order',
            'result': ('path-{}', 'status-{}'),
        },
        {
            'type': 'path-traj-{}',
            'name': 'path_ensemble-traj',
            'result': ('path-{}', 'status-{}', 'pathensemble-{}'),
        },
        {
            'type': 'path-energy',
            'name': 'path_ensemble-energy',
            'result': ('path-{}', 'status-{}'),
        },
    ]

    def __init__(self, ensembles, settings, controls):
        """Initialise the path simulation object.

        Parameters
        ----------
        ensembles : list of dicts
            Each contains:

            * `path_ensemble`: object like :py:class:`.PathEnsemble`
              This is used for storing results for the simulation. It
              is also used for defining the interfaces for this simulation.
            * `system`: object like :py:class:`.System`
              This is the system we are investigating.
            * `order_function`: object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `engine`: object like :py:class:`.EngineBase`
              This is the integrator that is used to propagate the
              system in time.
            * `rgen`: object like :py:class:`.RandomGenerator`
              This is the random generator to use in the ensemble.

        settings : dict
            This dictionary contains the settings for the simulation.
        controls: dict of parameters to set up and control the simulations
            It contains:

            * `steps`: int, optional
              The number of simulation steps to perform.
            * `startcycle`: int, optional
              The cycle we start the simulation on, useful for restarts.
            * `rgen`: object like :py:class:`.RandomGenerator`
              This object is the random generator to use in the simulation.

        """
        super().__init__(settings, controls)
        self.ensembles = ensembles
        self.settings = settings
        self.rgen = controls.get('rgen', create_random_generator())

        for key in self.required_settings:
            if key not in self.settings:
                logtxt = 'Missing required setting "{}" for simulation "{}"'
                logtxt = logtxt.format(key, self.name)
                logger.error(logtxt)
                raise ValueError(logtxt)
            self.settings[key] = settings[key]

        # Additional setup for shooting:
        for i, ensemble in enumerate(ensembles):
            ensemble['system'].potential_and_force()

            if self.settings['ensemble'][i]['tis']['sigma_v'] < 0.0:
                self.settings['ensemble'][i]['tis']['aimless'] = True
                logger.debug('%s: aimless is True', self.name)
            else:
                logger.debug('Path simulation: Creating sigma_v.')
                sigv = (self.settings['ensemble'][i]['tis']['sigma_v'] *
                        np.sqrt(ensemble['system'].particles.imass))
                logger.debug('Path simulation: sigma_v created and set.')
                self.settings['ensemble'][i]['tis']['sigma_v'] = sigv
                self.settings['ensemble'][i]['tis']['aimless'] = False
                logger.debug('Path simulation: aimless is False')

    def restart_info(self):
        """Return restart info.

        The restart info for the path simulation includes the state of
        the random number generator(s).

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['simulation']['rgen'] = self.rgen.get_state()

        # Here we store only the necessary info to initialize the
        # ensemble objects constructed in `pyretis.setup.createsimulation`
        # Note: these info are going to be stored in restart.pyretis
        if hasattr(self, 'ensembles'):
            info['ensemble'] = []
            for ens in self.ensembles:
                info['ensemble'].append(
                    {'restart': os.path.join(
                        ens['path_ensemble'].ensemble_name_simple,
                        'ensemble.restart')})

        return info

    def load_restart_info(self, info):
        """Load restart information.

        Note: This method load the info for the main simulation, the actual
              ensemble restart is done in initiate_restart.

        Parameters
        ----------
        info : dict
            The dictionary with the restart information, should be
            similar to the dict produced by :py:func:`.restart_info`.

        """
        super().load_restart_info(info)
        self.rgen = create_random_generator(info['simulation']['rgen'])

    def create_output_tasks(self, settings, progress=False):
        """Create output tasks for the simulation.

        This method will generate output tasks based on the tasks
        listed in :py:attr:`.simulation_output`.

        Parameters
        ----------
        settings : dict
            These are the simulation settings.
        progress : boolean
            For some simulations, the user may select to display a
            progress bar, we then need to disable the screen output.

        """
        logging.debug('Clearing output tasks & adding pre-defined ones')
        self.output_tasks = []
        for ensemble in self.ensembles:
            path_ensemble = ensemble['path_ensemble']
            directory = path_ensemble.directory['path_ensemble']
            idx = path_ensemble.ensemble_number
            logger.info('Creating output directories for path_ensemble %s',
                        path_ensemble.ensemble_name)
            for dir_name in path_ensemble.directories():
                msg_dir = make_dirs(dir_name)
                logger.info('%s', msg_dir)
            for task_dict in self.simulation_output:
                task_dict_ens = task_dict.copy()
                if 'result' in task_dict_ens:
                    task_dict_ens['result'] = \
                        [key.format(idx) for key in task_dict_ens['result']]
                task = task_from_settings(task_dict_ens, settings, directory,
                                          ensemble['engine'], progress)
                if task is not None:
                    logger.debug('Created output task:\n%s', task)
                    self.output_tasks.append(task)

    def write_restart(self, now=False):
        """Create a restart file.

        Parameters
        ----------
        now : boolean, optional
            If True, the output file will be written irrespective of the
            step number.

        """
        super().write_restart(now=now)
        if now or (self.restart_freq is not None and
                   self.cycle['stepno'] % self.restart_freq == 0):
            for ens, e_set in zip(self.ensembles, self.settings['ensemble']):
                write_ensemble_restart(ens, e_set)

    def initiate(self, settings):
        """Initialise the path simulation.

        Parameters
        ----------
        settings : dictionary
            The simulation settings.

        """
        init = initiate_path_simulation(self, settings)
        print_to_screen('')
        for i_ens, (accept, path, status, path_ensemble) in enumerate(init):
            print_to_screen(
                f'Found initial path for {path_ensemble.ensemble_name}:',
                level='success' if accept else 'warning',
            )
            for line in str(path).split('\n'):
                print_to_screen(f'- {line}')
            logger.info('Found initial path for %s',
                        path_ensemble.ensemble_name)
            logger.info('%s', path)
            print_to_screen('')
            idx = path_ensemble.ensemble_number
            path_ensemble_result = {
                f'pathensemble-{idx}': path_ensemble,
                f'path-{idx}': path,
                f'status-{idx}': status,
                'cycle': self.cycle,
                'system': self.system,
            }
            # If we are doing a restart, we do not print out at the
            # "restart" step as we assume that this is already
            # outputted in the "previous" simulation (the one
            # we restart from):
            if settings['initial-path']['method'] != 'restart':
                for task in self.output_tasks:
                    task.output(path_ensemble_result)
                write_ensemble_restart({'path_ensemble': path_ensemble},
                                       settings['ensemble'][i_ens])
            if self.soft_exit():
                return False
        return True

    def step(self):
        """Perform a TIS/RETIS/PPRETIS simulation step.

        Returns
        -------
        out : dict
            This list contains the results of the defined tasks.

        """
        sim_type = self.settings['simulation']['task'].lower()
        sim_map = {'explore': make_tis,
                   'tis': make_tis,
                   'pptis': make_tis,
                   'retis': make_retis_step,
                   'repptis': make_retis_step,
                   }

        prio_skip = priority_checker(self.ensembles, self.settings)
        if True not in prio_skip:
            self.cycle['step'] += 1
            self.cycle['stepno'] += 1
        msgtxt = f' {sim_type} step. Cycle {self.cycle["stepno"]}'
        logger.info(msgtxt)
        prepare = sim_map[sim_type]
        runner = prepare(self.ensembles, self.rgen,
                         self.settings, self.cycle['step'])
        results = {}
        for i_ens, res in enumerate(runner):
            if prio_skip[i_ens]:
                continue
            idx = res['ensemble_number']
            result = {'cycle': self.cycle}
            result[f'move-{idx}'] = res['mc-move']
            result[f'status-{idx}'] = res['status']
            result[f'path-{idx}'] = res['trial']
            result[f'accept-{idx}'] = res['accept']
            result[f'all-{idx}'] = res
            # This is to fix swaps needs
            idx_ens = i_ens if sim_type in {'pptis', 'tis', 'explore'} else idx
            result[f'pathensemble-{idx}'] = \
                self.ensembles[idx_ens]['path_ensemble']
            for task in self.output_tasks:
                task.output(result)
            results.update(result)
            if self.settings['ensemble'][idx_ens]['simulation'].get(
                    'remove_generate', True):
                logger.debug("Removing generate/ files:")
                logger.debug("self.ensembles[i_ens]['path_ensemble'] = %s",
                             self.ensembles[i_ens]['path_ensemble'])
                self.ensembles[i_ens]['path_ensemble'].clear_generate()
            if soft_partial_exit(self.settings['simulation']['exe_path']):
                self.cycle['endcycle'] = self.cycle['step']
                break
        return results

    def run(self):
        """Run a path simulation.

        The intended usage is for simulations where all tasks have
        been defined in :py:attr:`self.tasks`.

        Note
        ----
        This function will just run the tasks via executing
        :py:meth:`.step` In general, this is probably too generic for
        the simulation you want, if you are creating a custom simulation.
        Please consider customizing the :py:meth:`.run` (or the
        :py:meth:`.step`) method of your simulation class.

        Yields
        ------
        out : dict
            This dictionary contains the results from the simulation.

        """
        while not self.is_finished():
            result = self.step()
            self.write_restart()
            if self.soft_exit():
                yield result
                break
            yield result


class SimulationTIS(PathSimulation):
    """A TIS simulation.

    This class is used to define a TIS simulation where the goal is
    to calculate crossing probabilities for a single path ensemble.

    """

    required_settings = ('tis',)
    name = 'TIS simulation'
    simulation_type = 'tis'
    simulation_output = PathSimulation.simulation_output + [
        {
            'type': 'pathensemble-screen',
            'name': 'path_ensemble-screen',
            'result': ('pathensemble-{}',),
        },
    ]

    def __str__(self):
        """Just a small function to return some info about the simulation."""
        msg = ['TIS simulation']
        msg += ['Ensembles:']
        for ensemble in self.ensembles:
            path_ensemble = ensemble['path_ensemble']
            msg += [f'Path ensemble: {path_ensemble.ensemble_number}']
            msg += [(f'{path_ensemble.ensemble_name}: '
                     f'Interfaces: {path_ensemble.interfaces}')
                    ]
            msg += [f'Engine: {ensemble["engine"]}']

        nstep = self.cycle['endcycle'] - self.cycle['startcycle']
        msg += [f'Number of steps to do: {nstep}']
        return '\n'.join(msg)


class SimulationRETIS(PathSimulation):
    """A RETIS simulation.

    This class is used to define a RETIS simulation where the goal is
    to calculate crossing probabilities for several path ensembles.

    The attributes are documented in the parent class, please see:
    :py:class:`.PathSimulation`.
    """

    required_settings = ('retis',)
    name = 'RETIS simulation'
    simulation_type = 'retis'
    simulation_output = PathSimulation.simulation_output + [
        {
            'type': 'pathensemble-retis-screen',
            'name': 'path_ensemble-retis-screen',
            'result': ('pathensemble-{}',),
        },
    ]

    def __str__(self):
        """Just a small function to return some info about the simulation."""
        msg = ['RETIS simulation']
        msg += ['Ensembles:']
        for ensemble in self.ensembles:
            path_ensemble = ensemble['path_ensemble']
            msgtxt = (f'{path_ensemble.ensemble_name}: '
                      f'Interfaces: {path_ensemble.interfaces}')
            msg += [msgtxt]
        nstep = self.cycle['endcycle'] - self.cycle['startcycle']
        msg += [f'Number of steps to do: {nstep}']
        return '\n'.join(msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.

Here we compare a RETIS simulation of 250 steps to known results.
"""
import os
import sys
import tarfile
import colorama
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name


RESULTS = 'results'
RESULTS_TGZ = {
    '5.1.4': 'results-5.1.4.tgz',
}


def read_tarfile(tar_file, file_to_extract):
    """Extract and read the requested file from a .tgz archive.

    Parameters
    ----------
    tar_file : string
        Path to the .tgz file to open,
    file_to_extract : string
        The file we are going to extract.

    Returns
    -------
    out : list of strings
        The file we read.

    Note
    ----
    This will read the contents of the requested file into memory.

    """
    data = None
    with tarfile.open(tar_file, 'r:gz') as tar:
        for member in tar.getmembers():
            if member.isreg() and member.name == file_to_extract:
                tfile = tar.extractfile(member)
                data = tfile.read()
    return data


def find_traj_txt(rootdir):
    """Find all traj.txt files in the given rootdir.

    Here, we also sort these files according to the cycle number
    they claim to be produced at.

    """
    # Find all traj.txt files:
    files = []
    for root, _, filei in os.walk(rootdir):
        for i in filei:
            if i == 'traj.txt':
                filename = os.path.join(root, i)
                cycle = None
                with open(filename, 'r', encoding='utf-8') as infile:
                    line = infile.readline()
                    cycle = int(line.strip().split(':')[1].split(',')[0])
                if cycle is None:
                    raise ValueError('Could not find cyle for traj.txt')
                files.append([os.path.join(root, i), cycle])
    return sorted(files, key=lambda i: i[1])


def make_traj_file(root, ensemble):
    """Make a traj.txt file with paths.

    Here, we do not have the complete traj.txt file available and we thus
    recreate it using the stored accepted and rejected trajectories.

    Parameters
    ----------
    root : string
        The path to the directory containing the output.
    ensemble : integer
        The ensemble we are comparing for (0, 1, 2, ...)

    """
    ensemble_dir = generate_ensemble_name(ensemble)
    traj_dir = os.path.join(root, ensemble_dir, 'traj')
    files = find_traj_txt(traj_dir)
    target_traj = os.path.join(root, ensemble_dir, 'traj.txt')
    with open(target_traj, 'w', encoding='utf-8') as outfile:
        for (filename, _) in files:
            with open(filename, 'r', encoding='utf-8') as infile:
                outfile.write(infile.read())


def compare_files(settings, root, results_tgz):
    """Compare ouput files."""
    inter = settings['simulation']['interfaces']
    for i in range(len(inter)):
        for files in ('pathensemble.txt', 'order.txt', 'traj.txt'):
            if files == 'traj.txt':
                make_traj_file(root, i)
            result = compare_file_contents(root, i, files, results_tgz)
            if not result:
                print_to_screen('\t-> *Files differ!*', level='error')
                return False
    print_to_screen('All files are equal!', level='success')
    return True


def compare_file_contents(root, ensemble, file_name, results_tgz):
    """Compare files for a given ensemble inside a given directory.

    Parameters
    ----------
    root : string
        The directory where we will do the comparison.
    ensemble : integer
        The ensemble we are comparing for (0, 1, 2, ...)
    file_name : string
        The file we will compare.
    results_tgz : string
        The file containing the correct results we compare to.
        This is a tgz-archive.

    Returns
    -------
    result : boolean
        True if the files are equal, False otherwise.

    """
    tgz_file = os.path.join(root, results_tgz)
    ensemble_dir = generate_ensemble_name(ensemble)
    file2 = os.path.join(RESULTS, ensemble_dir, file_name)
    data2 = read_tarfile(tgz_file, file2)
    file1 = os.path.join(root, ensemble_dir, file_name)
    print_to_screen(f'Comparing for: {file1}')
    return open(file1, 'rb').read() == data2


def main(args):
    """Run the comparison."""
    try:
        gmx_version = args[1].split()[-1]
    except IndexError:
        gmx_version = '5.1.4'
    print_to_screen(f'GROMACS version family: {gmx_version}',
                    level='info')
    correct = RESULTS_TGZ.get(gmx_version, None)
    if correct is None:
        print_to_screen(
            f'GROMACS version {gmx_version} is not supported ' +
            'for result comparison.'
        )
        print_to_screen('The supported GROMACS versions are:')
        for key in RESULTS_TGZ:
            print_to_screen(f'- {key}')
        sys.exit(0)

    print_to_screen(f'Using results from: {correct}', level='info')

    for dirname in ('gromacs1', 'gromacs2'):
        sets = parse_settings_file(os.path.join(dirname, 'retis.rst'))
        head = f'Comparing files for: {dirname}'
        print_to_screen(f'\n{head}', level='message')
        print_to_screen('=' * len(head), level='message')
        compare_ok = compare_files(sets, dirname, correct)
        if not compare_ok:
            sys.exit(1)


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main(sys.argv)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the CP2K io module."""
import logging
import unittest
import tempfile
import os
import numpy as np
from pyretis.inout.formats.cp2k import (
    read_cp2k_input,
    dfs_print,
    set_parents,
    read_cp2k_energy,
    read_cp2k_restart,
    read_cp2k_box,
    update_node,
    remove_node,
)


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))

BOX_TEMPLATE = """&FORCE_EVAL
  &SUBSYS
    &CELL
      PERIODIC {}
      ABC 5.0 5.0 5.0
      ALPHA_BETA_GAMMA 90.0 90.0 90.0
    &END CELL
  &END SUBSYS
&END FORCE_EVAL"""
BOX_TEMPLATE2 = """&FORCE_EVAL
  &SUBSYS
    &CELL
      ABC 5.0 5.0 5.0
      ALPHA_BETA_GAMMA 90.0 90.0 90.0
    &END CELL
  &END SUBSYS
&END FORCE_EVAL"""
BOX_TEMPLATE3 = """&FORCE_EVAL
  &SUBSYS
  &END SUBSYS
&END FORCE_EVAL"""


class CP2KIOTest(unittest.TestCase):
    """Test CP2KIO."""

    def test_read_cp2k(self):
        """Test that we can read CP2K input files."""
        infile = os.path.join(HERE, 'cp2k.inp')
        nodes = read_cp2k_input(infile)
        buff = []
        for _, node in enumerate(nodes):
            visited = set()
            out = dfs_print(node, visited)
            for line in out:
                buff.append(line.strip())
        with open(infile, 'r') as inputfile:
            buff2 = [i.strip() for i in inputfile]
            # Test that we find all lines in the original input file:
            for line in buff:
                self.assertTrue(line in buff2)
            # Test that we read everything:
            for line in buff2:
                if not line:
                    continue
                self.assertTrue(line in buff)

    def test_set_parents_read_cp2k(self):
        """Test that we can construct parents for sections."""
        infile = os.path.join(HERE, 'cp2k.inp')
        nodes = read_cp2k_input(infile)
        node_ref = set_parents(nodes)
        ref = 'FORCE_EVAL->DFT->XC->XC_FUNCTIONAL'
        node = node_ref[ref]
        self.assertEqual(node.title, 'XC_FUNCTIONAL')
        self.assertEqual(node.level, 3)
        for i, j in zip(node.settings, ['PADE']):
            self.assertEqual(i, j)
        for i, j in zip(node.parents, ['FORCE_EVAL', 'DFT', 'XC',
                                       'XC_FUNCTIONAL']):
            self.assertEqual(i, j)

    def test_read_energy(self):
        """Test that we can read cp2k energy output."""
        infile = os.path.join(HERE, 'cp2k.ener')
        energy = read_cp2k_energy(infile)
        for key in ('ekin', 'vpot', 'temp', 'etot'):
            self.assertTrue(key in energy)
        kin_correct = np.array([0.004750225, 0.008063503, 0.018933585,
                                0.033672265, 0.048363565, 0.060298562])
        pot_correct = np.array([-15.906282000, -15.920520150, -15.931604156,
                                -15.946600436, -15.961490491, -15.973407929])
        self.assertTrue(np.allclose(kin_correct, energy['ekin']))
        self.assertTrue(np.allclose(pot_correct, energy['vpot']))

    def test_read_energy_missing(self):
        """Test that we can read cp2k energy output with missing columns."""
        infile = os.path.join(HERE, 'cp2k.ener2')
        energy = read_cp2k_energy(infile)
        for key in ('ekin', ):
            self.assertTrue(key in energy)
        for key in ('vpot', 'temp', 'etot'):
            self.assertFalse(key in energy)
        kin_correct = np.array([0.004750225, 0.008063503, 0.018933585,
                                0.033672265, 0.048363565, 0.060298562])
        self.assertTrue(np.allclose(kin_correct, energy['ekin']))

    def test_read_restart(self):
        """Test that we can read cp2k restart files."""
        infile = os.path.join(HERE, 'cp2k.restart')
        atoms, pos, vel, box, periodic = read_cp2k_restart(infile)
        correct = {
            'atoms': ['H', 'N', 'C'],
            'pos': np.array([
                [2.0037295984360898E-01, 1.3196012905205206E+00,
                 -8.4772218235716257E-01],
                [1.2469580920891239E+00, 5.6235164153925576E-01,
                 -1.5756836410607700E-01],
                [-1.5843560585520555E-01, 3.9769694640060993E-01,
                 2.4037061887112918E-01],
                ]),
            'vel': np.array([
                [-1.3106795516477289E-03, -1.7627790939445680E-03,
                 -4.3827056898331971E-04],
                [-7.7711186323380392E-04, -4.0307772046066932E-04,
                 4.6944742420092559E-04],
                [1.0163250974302026E-03, 6.1937600463006590E-04,
                 -5.3549195000426192E-04],
                ]),
            'box': [5., 5., 5.],
            'periodic': [True, True, True],
        }
        for i, j in zip(atoms, correct['atoms']):
            self.assertEqual(i, j)
        self.assertTrue(np.allclose(pos, correct['pos']))
        self.assertTrue(np.allclose(vel, correct['vel']))
        for i, j in zip(periodic, correct['periodic']):
            self.assertEqual(i, j)
        for i, j in zip(box, correct['box']):
            self.assertAlmostEqual(i, j)

    def test_read_box(self):
        """Test that we can read the box."""
        test_data = [
            {'file': 'box1.inp', 'box': [5., 5., 5.],
             'periodic': [True, True, True]},
            {'file': 'box2.inp', 'box': [1., 5., 9., 4., 7., 2., 8., 3., 6.],
             'periodic': [True, True, True]},
            {'file': 'box3.inp',
             'box': [2.7, 4.47637, 1.97371, 1.99301, 0.25593,
                     0.0, -0.67006, 0.0, 0.0], 'periodic': [True, True, True]},
            {'file': 'box4.inp', 'box': [5., 5., 5.],
             'periodic': [True, True, True]},
            {'file': 'box5.inp', 'box': None,
             'periodic': [True, True, True]},
        ]
        for data in test_data:
            infile = os.path.join(HERE, data['file'])
            box, periodic = read_cp2k_box(infile)
            if box is None:
                self.assertEqual(box, data['box'])
            else:
                for i, j in zip(box, data['box']):
                    self.assertAlmostEqual(i, j, places=4)
            for i, j in zip(periodic, data['periodic']):
                self.assertEqual(i, j)

    def test_read_box_periodic(self):
        """Test that we can read different periodic settings."""
        correct = {
            'X': [True, False, False],
            'XY': [True, True, False],
            'XYZ': [True, True, True],
            'XZ': [True, False, True],
            'Y': [False, True, False],
            'YZ': [False, True, True],
            'Z': [False, False, True],
        }
        for key, val in correct.items():
            txt = BOX_TEMPLATE.format(key)
            with tempfile.NamedTemporaryFile() as temp:
                temp.write(txt.encode('utf-8'))
                temp.flush()
                _, periodic = read_cp2k_box(temp.name)
                for i, j in zip(periodic, val):
                    self.assertEqual(i, j)
        # Test also when we are missing the input:
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(BOX_TEMPLATE2.encode('utf-8'))
            temp.flush()
            _, periodic = read_cp2k_box(temp.name)
            for i, j in zip(periodic, [True, True, True]):
                self.assertEqual(i, j)
        with tempfile.NamedTemporaryFile() as temp:
            temp.write(BOX_TEMPLATE3.encode('utf-8'))
            temp.flush()
            _, periodic = read_cp2k_box(temp.name)
            for i, j in zip(periodic, [True, True, True]):
                self.assertEqual(i, j)

    def test_update_node(self):
        """Test that we can update nodes."""
        infile = os.path.join(HERE, 'cp2k.inp')
        nodes = read_cp2k_input(infile)
        node_ref = set_parents(nodes)
        target = 'MOTION->MD'
        data = {'STEPS': 987}
        self.assertEqual(node_ref[target].data[0], 'STEPS 123')
        update_node(target, None, data, node_ref, nodes, replace=False)
        self.assertEqual(node_ref[target].data[0], 'STEPS 987')
        data = ['STEPS 666']
        update_node(target, None, data, node_ref, nodes, replace=True)
        self.assertEqual(node_ref[target].data[0], 'STEPS 666')
        data = {'EXTRA': 'yes', 'EXTRAEXTRA': None}
        update_node(target, None, data, node_ref, nodes, replace=False)
        for i in ('STEPS 666', 'EXTRA yes', 'EXTRAEXTRA'):
            self.assertTrue(i in node_ref[target].data)
        settings = ['one', 'two']
        update_node(target, settings, [], node_ref, nodes, replace=False)
        for i, j in zip(settings, node_ref[target].settings):
            self.assertEqual(i, j)
        settings = ['three']
        update_node(target, settings, [], node_ref, nodes, replace=True)
        for i, j in zip(settings, node_ref[target].settings):
            self.assertEqual(i, j)
        target = 'MOTION->STUFF'
        settings = ['nope']
        data = ['setting1 100']
        update_node(target, settings, data, node_ref, nodes, replace=False)
        self.assertEqual(node_ref[target].data[0], 'setting1 100')
        self.assertEqual(node_ref[target].settings[0], 'nope')
        self.assertEqual(len(node_ref), 26)
        target = 'SOMETHING->WHATEVER'
        settings = ['nope']
        data = ['whatever']
        update_node(target, settings, data, node_ref, nodes, replace=False)
        self.assertEqual(len(node_ref), 28)
        self.assertTrue('SOMETHING->WHATEVER' in node_ref)

    def test_remove_node(self):
        """Test removal of nodes."""
        target = 'MOTION'
        infile = os.path.join(HERE, 'cp2k.inp')
        nodes = read_cp2k_input(infile)
        node_ref = set_parents(nodes)
        self.assertTrue('MOTION' in node_ref)
        remove_node(target, node_ref, nodes)
        self.assertFalse('MOTION' in node_ref)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the common methods in pyretis.pyvisa.common."""
import colorama
import os
import logging
import unittest
import numpy as np
import tempfile
import shutil
from unittest import mock
from io import StringIO
from pyretis.pyvisa.common import (
    try_data_shift,
    where_from_to,
    read_traj_txt_file,
    find_rst_file,
    _get_trjs,
    find_data,
    recalculate_all
)
logging.disable(logging.CRITICAL)

HERE = os.path.abspath(os.path.dirname(__file__))
color = colorama.Fore.CYAN


def del_list_ind(data_list, to_delete):
    """Function deleting indices in list."""
    for i in reversed(to_delete):
        del data_list[i]


class TestMethods(unittest.TestCase):
    """Test some of the methods from pyretis.pyvisa.common."""

    def test_try_data_shift(self):
        """Test to try_data_shift and shift_data functions."""
        n = 100  # number of data points
        # The Correct data, normal distr around origo
        # with a slope m=-1/3
        x, y = [], []
        np.random.seed(n)
        for i in range(n):
            r1 = np.random.normal()
            r2 = 3*np.random.normal()
            x.append(r1-r2)
            y.append(r1+r2)
        # The shifted data, to be corrected
        avx, avy = np.average(x), np.average(y)
        sx, sy = [], []
        for i in range(n):
            if x[i] > avx:
                sx.append(-90+x[i])
            else:
                sx.append(x[i])
            if y[i] > avy:
                sy.append(-90+y[i])
            else:
                sy.append(y[i])
        # Evaluate correct data, should do nothing:
        nx, ny = try_data_shift(x, y, False)
        self.assertEqual(x, nx)
        self.assertEqual(y, ny)
        # Evaluate x-shifted data:
        nx, ny = try_data_shift(sx, y, False)
        self.assertEqual(ny, y)
        self.assertFalse(np.average(sx) == np.average(nx))
        # Evaluate y-shifted data:
        nx, ny = try_data_shift(x, sy, False)
        self.assertEqual(x, nx)
        self.assertFalse(np.average(sy) == np.average(ny))
        # Evaluate x&y-shifted data:
        nx, ny = try_data_shift(sx, sy, False)
        self.assertFalse(np.average(sx) == np.average(nx))
        self.assertFalse(np.average(sy) == np.average(ny))
        # Evaluate x&y-shifted data, but with op1 flag to hold:
        nx, ny = try_data_shift(sx, sy, True)
        self.assertEqual(sx, nx)
        self.assertFalse(np.average(sy) == np.average(ny))

    def test_where_to_from(self):
        """Test to get the right L and R and * trj start and end labels."""
        trj = [1, 2, 3, 4, 5, 6, 4]
        int_as, int_bs = [0, 2, 2, 0, 5, 3], [7, 7, 3, 0, 5, 2]
        results = [('*', '*'), ('L', '*'), ('L', 'R'),
                   ('R', 'R'), ('L', 'L'), ('L', 'R')]

        # For a certain trj, we test different interfaces.
        for int_a, int_b, result in zip(int_as, int_bs,  results):
            start, end = where_from_to(trj, int_a, int_b)
            self.assertEqual((start, end), result)

        # Let's check the zero ensemble.
        trjs = [[0, 2], [6, 5], [1, 6]]
        results = [('*', '*'), ('R', 'R'), ('*', 'R')]

        for trj, result in zip(trjs, results):
            start, end = where_from_to(trj, int_a=3)
            self.assertEqual((start, end), result)

    def test_find_rst_file(self):
        """Test for finding the rst file."""
        start_path = os.path.join(HERE, 'test_simulation_dir/000')
        os.chdir(start_path)
        rst_path = os.path.join(HERE, 'test_simulation_dir/input.rst')
        final_path = find_rst_file(start_path)
        self.assertEqual(rst_path, final_path)
        final_path = find_rst_file('/')
        self.assertEqual('/', '/')

    def test_read_traj_txt_file(self):
        """Test for reading a traj.txt file."""
        txt_file = os.path.join(HERE, 'test_simulation_dir/traj.txt')
        files = read_traj_txt_file(txt_file)
        self.assertEqual(len(files.keys()), 2)

    def test_get_trjs(self):
        """Test for _get_trjs."""
        dir_file = os.path.join(HERE, 'test_simulation_dir')
        trj_list = _get_trjs(dir_file)
        found = sum([1 for trj in trj_list if 'traj.' in trj])
        self.assertGreater(found, 0)

    def test_find_data(self):
        """Test for find_data."""
        ass = find_data(HERE, data=HERE + '/test_simulation_dir/traj.txt')
        self.assertIn('traj.', ass['000']['traj']['0']['traj'][0])

        source = os.path.join(HERE, 'test_simulation_dir/003/order.txt')
        tmp = os.path.join(HERE, 'test_simulation_dir/003/banana')
        os.rename(source, tmp)
        ass = find_data(os.path.join(HERE, 'test_simulation_dir'))
        target = os.path.join(HERE, 'test_simulation_dir',
                              '003', 'traj', 'traj-acc',
                              '0', 'traj', 'traj.xyz')
        self.assertEqual(ass['003']['traj']['0']['traj'][0], target)
        os.rename(tmp, source)

        ass = find_data(HERE, data=HERE + '/test_simulation_dir/traj.txt')
        self.assertIn('test_simulation_dir/traj.txt',
                      ass['000']['traj']['0']['traj'][0])

    def test_recalculate_all(self):
        """Test for recalculate_all."""
        with tempfile.TemporaryDirectory() as tempdir:
            tmp_dir = os.path.join(tempdir, 'test_simulation_dir')
            shutil.copytree(HERE + '/test_simulation_dir', tmp_dir)
            with mock.patch('sys.stdout', new=StringIO()):
                self.assertFalse(recalculate_all(tmp_dir, 'input.rst'))
                file_new = os.path.join(tmp_dir, 'input.rst')
                os.system(f"sed -i s/'class = Position'/' '/ {file_new}")
                self.assertTrue(recalculate_all(tmp_dir, 'input.rst'))
                self.assertFalse(recalculate_all(tmp_dir, 'input.rst',
                                                 ensemble_names=[]))


if __name__ == '__main__':
    unittest.main()
